#
# Copyright 2017 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
import warnings

import empyrical as ep
from pandas.tseries.offsets import BDay
from scipy import stats
from statsmodels.regression.linear_model import OLS
from statsmodels.tools.tools import add_constant
from . import utils


def factor_information_coefficient(factor_data,
                                   group_adjust=False,
                                   by_group=False):
    """
    Computes the Spearman Rank Correlation based Information Coefficient (IC)
    between factor values and N period forward returns for each period in
    the factor index.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    group_adjust : bool
        Demean forward returns by group before computing IC.
    by_group : bool
        If True, compute period wise IC separately for each group.

    Returns
    -------
    ic : pd.DataFrame
        Spearman Rank correlation between factor and
        provided forward returns.
    """

    def src_ic(group):
        f = group['factor']
        _ic = group[utils.get_forward_returns_columns(factor_data.columns)] \
            .apply(lambda x: stats.spearmanr(x, f)[0])
        return _ic

    factor_data = factor_data.copy()

    grouper = [factor_data.index.get_level_values('date')]

    if group_adjust:
        factor_data = utils.demean_forward_returns(factor_data,
                                                   grouper + ['group'])
    if by_group:
        grouper.append('group')

    ic = factor_data.groupby(grouper).apply(src_ic)

    return ic


def mean_information_coefficient(factor_data,
                                 group_adjust=False,
                                 by_group=False,
                                 by_time=None):
    """
    Get the mean information coefficient of specified groups.
    Answers questions like:
    What is the mean IC for each month?
    What is the mean IC for each group for our whole timerange?
    What is the mean IC for for each group, each week?

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    group_adjust : bool
        Demean forward returns by group before computing IC.
    by_group : bool
        If True, take the mean IC for each group.
    by_time : str (pd time_rule), optional
        Time window to use when taking mean IC.
        See http://pandas.pydata.org/pandas-docs/stable/timeseries.html
        for available options.

    Returns
    -------
    ic : pd.DataFrame
        Mean Spearman Rank correlation between factor and provided
        forward price movement windows.
    """

    ic = factor_information_coefficient(factor_data, group_adjust, by_group)

    grouper = []
    if by_time is not None:
        grouper.append(pd.Grouper(freq=by_time))
    if by_group:
        grouper.append('group')

    if len(grouper) == 0:
        ic = ic.mean()

    else:
        ic = (ic.reset_index().set_index('date').groupby(grouper).mean())

    return ic


def factor_weights(factor_data,
                   demeaned=True,
                   group_adjust=False,
                   equal_weight=False):
    """
    Computes asset weights by factor values and dividing by the sum of their
    absolute value (achieving gross leverage of 1). Positive factor values will
    results in positive weights and negative values in negative weights.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    demeaned : bool
        Should this computation happen on a long short portfolio? if True,
        weights are computed by demeaning factor values and dividing by the sum
        of their absolute value (achieving gross leverage of 1). The sum of
        positive weights will be the same as the negative weights (absolute
        value), suitable for a dollar neutral long-short portfolio
    group_adjust : bool
        Should this computation happen on a group neutral portfolio? If True,
        compute group neutral weights: each group will weight the same and
        if 'demeaned' is enabled the factor values demeaning will occur on the
        group level.
    equal_weight : bool, optional
        if True the assets will be equal-weighted instead of factor-weighted
        If demeaned is True then the factor universe will be split in two
        equal sized groups, top assets with positive weights and bottom assets
        with negative weights

    Returns
    -------
    returns : pd.Series
        Assets weighted by factor value.
    """

    def to_weights(group, _demeaned, _equal_weight):

        if _equal_weight:
            group = group.copy()

            if _demeaned:
                # top assets positive weights, bottom ones negative
                group = group - group.median()

            negative_mask = group < 0
            group[negative_mask] = -1.0
            positive_mask = group > 0
            group[positive_mask] = 1.0

            if _demeaned:
                # positive weights must equal negative weights
                if negative_mask.any():
                    group[negative_mask] /= negative_mask.sum()
                if positive_mask.any():
                    group[positive_mask] /= positive_mask.sum()

        elif _demeaned:
            group = group - group.mean()

        return group / group.abs().sum()

    grouper = [factor_data.index.get_level_values('date')]
    if group_adjust:
        grouper.append('group')

    weights = factor_data.groupby(grouper)['factor'] \
        .apply(to_weights, demeaned, equal_weight)

    if group_adjust:
        weights = weights.groupby(level='date').apply(to_weights, False, False)

    return weights


def factor_returns(factor_data,
                   demeaned=True,
                   group_adjust=False,
                   equal_weight=False,
                   by_asset=False):
    """
    Computes period wise returns for portfolio weighted by factor
    values.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    demeaned : bool
        Control how to build factor weights
        -- see performance.factor_weights for a full explanation
    group_adjust : bool
        Control how to build factor weights
        -- see performance.factor_weights for a full explanation
    equal_weight : bool, optional
        Control how to build factor weights
        -- see performance.factor_weights for a full explanation
    by_asset: bool, optional
        If True, returns are reported separately for each esset.

    Returns
    -------
    returns : pd.DataFrame
        Period wise factor returns
    """

    weights = \
        factor_weights(factor_data, demeaned, group_adjust, equal_weight)

    weighted_returns = \
        factor_data[utils.get_forward_returns_columns(factor_data.columns)] \
        .multiply(weights, axis=0)

    if by_asset:
        returns = weighted_returns
    else:
        returns = weighted_returns.groupby(level='date').sum()

    return returns


def factor_alpha_beta(factor_data,
                      returns=None,
                      demeaned=True,
                      group_adjust=False,
                      equal_weight=False):
    """
    Compute the alpha (excess returns), alpha t-stat (alpha significance),
    and beta (market exposure) of a factor. A regression is run with
    the period wise factor universe mean return as the independent variable
    and mean period wise return from a portfolio weighted by factor values
    as the dependent variable.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    returns : pd.DataFrame, optional
        Period wise factor returns. If this is None then it will be computed
        with 'factor_returns' function and the passed flags: 'demeaned',
        'group_adjust', 'equal_weight'
    demeaned : bool
        Control how to build factor returns used for alpha/beta computation
        -- see performance.factor_return for a full explanation
    group_adjust : bool
        Control how to build factor returns used for alpha/beta computation
        -- see performance.factor_return for a full explanation
    equal_weight : bool, optional
        Control how to build factor returns used for alpha/beta computation
        -- see performance.factor_return for a full explanation

    Returns
    -------
    alpha_beta : pd.Series
        A list containing the alpha, beta, a t-stat(alpha)
        for the given factor and forward returns.
    """

    if returns is None:
        returns = \
            factor_returns(factor_data, demeaned, group_adjust, equal_weight)

    universe_ret = factor_data.groupby(level='date')[
        utils.get_forward_returns_columns(factor_data.columns)] \
        .mean().loc[returns.index]

    if isinstance(returns, pd.Series):
        returns.name = universe_ret.columns.values[0]
        returns = pd.DataFrame(returns)

    alpha_beta = pd.DataFrame()
    for period in returns.columns.values:
        x = universe_ret[period].values
        y = returns[period].values
        x = add_constant(x)

        reg_fit = OLS(y, x).fit()
        try:
            alpha, beta = reg_fit.params
        except ValueError:
            alpha_beta.loc['Ann. alpha', period] = np.nan
            alpha_beta.loc['beta', period] = np.nan
        else:
            freq_adjust = pd.Timedelta('252Days') / pd.Timedelta(period)

            alpha_beta.loc['Ann. alpha', period] = \
                (1 + alpha) ** freq_adjust - 1
            alpha_beta.loc['beta', period] = beta

    return alpha_beta


def cumulative_returns(returns):
    """
    Computes cumulative returns from simple daily returns.

    Parameters
    ----------
    returns: pd.Series
        pd.Series containing daily factor returns (i.e. '1D' returns).

    Returns
    -------
    Cumulative returns series : pd.Series
        Example:
            2015-01-05   1.001310
            2015-01-06   1.000805
            2015-01-07   1.001092
            2015-01-08   0.999200
    """

    return ep.cum_returns(returns, starting_value=1)


def positions(weights, period, freq=None):
    """
    Builds net position values time series, the portfolio percentage invested
    in each position.

    Parameters
    ----------
    weights: pd.Series
        pd.Series containing factor weights, the index contains timestamps at
        which the trades are computed and the values correspond to assets
        weights
        - see factor_weights for more details
    period: pandas.Timedelta or string
        Assets holding period (1 day, 2 mins, 3 hours etc). It can be a
        Timedelta or a string in the format accepted by Timedelta constructor
        ('1 days', '1D', '30m', '3h', '1D1h', etc)
    freq : pandas DateOffset, optional
        Used to specify a particular trading calendar. If not present
        weights.index.freq will be used

    Returns
    -------
    pd.DataFrame
        Assets positions series, datetime on index, assets on columns.
        Example:
            index                 'AAPL'         'MSFT'          cash
            2004-01-09 10:30:00   13939.3800     -14012.9930     711.5585
            2004-01-09 15:30:00       0.00       -16012.9930     411.5585
            2004-01-12 10:30:00   14492.6300     -14624.8700       0.0
            2004-01-12 15:30:00   14874.5400     -15841.2500       0.0
            2004-01-13 10:30:00   -13853.2800    13653.6400      -43.6375
    """

    weights = weights.unstack()

    if not isinstance(period, pd.Timedelta):
        period = pd.Timedelta(period)

    if freq is None:
        freq = weights.index.freq

    if freq is None:
        freq = BDay()
        warnings.warn("'freq' not set, using business day calendar",
                      UserWarning)

    #
    # weights index contains factor computation timestamps, then add returns
    # timestamps too (factor timestamps + period) and save them to 'full_idx'
    # 'full_idx' index will contain an entry for each point in time the weights
    # change and hence they have to be re-computed
    #
    trades_idx = weights.index.copy()
    returns_idx = utils.add_custom_calendar_timedelta(trades_idx, period, freq)
    weights_idx = trades_idx.union(returns_idx)

    #
    # Compute portfolio weights for each point in time contained in the index
    #
    portfolio_weights = pd.DataFrame(index=weights_idx,
                                     columns=weights.columns)
    active_weights = []

    for curr_time in weights_idx:

        #
        # fetch new weights that become available at curr_time and store them
        # in active weights
        #
        if curr_time in weights.index:
            assets_weights = weights.loc[curr_time]
            expire_ts = utils.add_custom_calendar_timedelta(curr_time,
                                                            period, freq)
            active_weights.append((expire_ts, assets_weights))

        #
        # remove expired entry in active_weights (older than 'period')
        #
        if active_weights:
            expire_ts, assets_weights = active_weights[0]
            if expire_ts <= curr_time:
                active_weights.pop(0)

        if not active_weights:
            continue
        #
        # Compute total weights for curr_time and store them
        #
        tot_weights = [w for (ts, w) in active_weights]
        tot_weights = pd.concat(tot_weights, axis=1)
        tot_weights = tot_weights.sum(axis=1)
        tot_weights /= tot_weights.abs().sum()

        portfolio_weights.loc[curr_time] = tot_weights

    return portfolio_weights.fillna(0)


def mean_return_by_quantile(factor_data,
                            by_date=False,
                            by_group=False,
                            demeaned=True,
                            group_adjust=False):
    """
    Computes mean returns for factor quantiles across
    provided forward returns columns.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    by_date : bool
        If True, compute quantile bucket returns separately for each date.
    by_group : bool
        If True, compute quantile bucket returns separately for each group.
    demeaned : bool
        Compute demeaned mean returns (long short portfolio)
    group_adjust : bool
        Returns demeaning will occur on the group level.

    Returns
    -------
    mean_ret : pd.DataFrame
        Mean period wise returns by specified factor quantile.
    std_error_ret : pd.DataFrame
        Standard error of returns by specified quantile.
    """

    if group_adjust:
        grouper = [factor_data.index.get_level_values('date')] + ['group']
        factor_data = utils.demean_forward_returns(factor_data, grouper)
    elif demeaned:
        factor_data = utils.demean_forward_returns(factor_data)
    else:
        factor_data = factor_data.copy()

    grouper = ['factor_quantile', factor_data.index.get_level_values('date')]

    if by_group:
        grouper.append('group')

    group_stats = factor_data.groupby(grouper)[
        utils.get_forward_returns_columns(factor_data.columns)] \
        .agg(['mean', 'std', 'count'])

    mean_ret = group_stats.T.xs('mean', level=1).T

    if not by_date:
        grouper = [mean_ret.index.get_level_values('factor_quantile')]
        if by_group:
            grouper.append(mean_ret.index.get_level_values('group'))
        group_stats = mean_ret.groupby(grouper)\
            .agg(['mean', 'std', 'count'])
        mean_ret = group_stats.T.xs('mean', level=1).T

    std_error_ret = group_stats.T.xs('std', level=1).T \
        / np.sqrt(group_stats.T.xs('count', level=1).T)

    return mean_ret, std_error_ret


def compute_mean_returns_spread(mean_returns,
                                upper_quant,
                                lower_quant,
                                std_err=None):
    """
    Computes the difference between the mean returns of
    two quantiles. Optionally, computes the standard error
    of this difference.

    Parameters
    ----------
    mean_returns : pd.DataFrame
        DataFrame of mean period wise returns by quantile.
        MultiIndex containing date and quantile.
        See mean_return_by_quantile.
    upper_quant : int
        Quantile of mean return from which we
        wish to subtract lower quantile mean return.
    lower_quant : int
        Quantile of mean return we wish to subtract
        from upper quantile mean return.
    std_err : pd.DataFrame, optional
        Period wise standard error in mean return by quantile.
        Takes the same form as mean_returns.

    Returns
    -------
    mean_return_difference : pd.Series
        Period wise difference in quantile returns.
    joint_std_err : pd.Series
        Period wise standard error of the difference in quantile returns.
        if std_err is None, this will be None
    """

    mean_return_difference = mean_returns.xs(upper_quant,
                                             level='factor_quantile') \
        - mean_returns.xs(lower_quant, level='factor_quantile')

    if std_err is None:
        joint_std_err = None
    else:
        std1 = std_err.xs(upper_quant, level='factor_quantile')
        std2 = std_err.xs(lower_quant, level='factor_quantile')
        joint_std_err = np.sqrt(std1**2 + std2**2)

    return mean_return_difference, joint_std_err


def quantile_turnover(quantile_factor, quantile, period=1):
    """
    Computes the proportion of names in a factor quantile that were
    not in that quantile in the previous period.

    Parameters
    ----------
    quantile_factor : pd.Series
        DataFrame with date, asset and factor quantile.
    quantile : int
        Quantile on which to perform turnover analysis.
    period: int, optional
        Number of days over which to calculate the turnover.

    Returns
    -------
    quant_turnover : pd.Series
        Period by period turnover for that quantile.
    """

    quant_names = quantile_factor[quantile_factor == quantile]
    quant_name_sets = quant_names.groupby(level=['date']).apply(
        lambda x: set(x.index.get_level_values('asset')))

    name_shifted = quant_name_sets.shift(period)

    new_names = (quant_name_sets - name_shifted).dropna()
    quant_turnover = new_names.apply(
        lambda x: len(x)) / quant_name_sets.apply(lambda x: len(x))
    quant_turnover.name = quantile
    return quant_turnover


def factor_rank_autocorrelation(factor_data, period=1):
    """
    Computes autocorrelation of mean factor ranks in specified time spans.
    We must compare period to period factor ranks rather than factor values
    to account for systematic shifts in the factor values of all names or names
    within a group. This metric is useful for measuring the turnover of a
    factor. If the value of a factor for each name changes randomly from period
    to period, we'd expect an autocorrelation of 0.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    period: int, optional
        Number of days over which to calculate the turnover.

    Returns
    -------
    autocorr : pd.Series
        Rolling 1 period (defined by time_rule) autocorrelation of
        factor values.
    """
    grouper = [factor_data.index.get_level_values('date')]

    ranks = factor_data.groupby(grouper)['factor'].rank()

    asset_factor_rank = ranks.reset_index().pivot(index='date',
                                                  columns='asset',
                                                  values='factor')

    asset_shifted = asset_factor_rank.shift(period)

    autocorr = asset_factor_rank.corrwith(asset_shifted, axis=1)
    autocorr.name = period
    return autocorr


def common_start_returns(factor,
                         returns,
                         before,
                         after,
                         cumulative=False,
                         mean_by_date=False,
                         demean_by=None):
    """
    A date and equity pair is extracted from each index row in the factor
    dataframe and for each of these pairs a return series is built starting
    from 'before' the date and ending 'after' the date specified in the pair.
    All those returns series are then aligned to a common index (-before to
    after) and returned as a single DataFrame

    Parameters
    ----------
    factor : pd.DataFrame
        DataFrame with at least date and equity as index, the columns are
        irrelevant
    returns : pd.DataFrame
        A wide form Pandas DataFrame indexed by date with assets in the
        columns. Returns data should span the factor analysis time period
        plus/minus an additional buffer window corresponding to after/before
        period parameters.
    before:
        How many returns to load before factor date
    after:
        How many returns to load after factor date
    cumulative: bool, optional
        Whether or not the given returns are cumulative. If False the given
        returns are assumed to be daily.
    mean_by_date: bool, optional
        If True, compute mean returns for each date and return that
        instead of a return series for each asset
    demean_by: pd.DataFrame, optional
        DataFrame with at least date and equity as index, the columns are
        irrelevant. For each date a list of equities is extracted from
        'demean_by' index and used as universe to compute demeaned mean
        returns (long short portfolio)

    Returns
    -------
    aligned_returns : pd.DataFrame
        Dataframe containing returns series for each factor aligned to the same
        index: -before to after
    """
    if not cumulative:
        returns = returns.apply(cumulative_returns, axis=0)

    all_returns = []

    for timestamp, df in factor.groupby(level='date'):

        equities = df.index.get_level_values('asset')

        try:
            day_zero_index = returns.index.get_loc(timestamp)
        except KeyError:
            continue

        starting_index = max(day_zero_index - before, 0)
        ending_index = min(day_zero_index + after + 1,
                           len(returns.index))

        equities_slice = set(equities)
        if demean_by is not None:
            demean_equities = demean_by.loc[timestamp] \
                .index.get_level_values('asset')
            equities_slice |= set(demean_equities)

        series = returns.loc[returns.index[starting_index:ending_index],
                             equities_slice]
        series.index = range(starting_index - day_zero_index,
                             ending_index - day_zero_index)

        if demean_by is not None:
            mean = series.loc[:, demean_equities].mean(axis=1)
            series = series.loc[:, equities]
            series = series.sub(mean, axis=0)

        if mean_by_date:
            series = series.mean(axis=1)

        all_returns.append(series)

    return pd.concat(all_returns, axis=1)


def average_cumulative_return_by_quantile(factor_data,
                                          returns,
                                          periods_before=10,
                                          periods_after=15,
                                          demeaned=True,
                                          group_adjust=False,
                                          by_group=False):
    """
    Plots average cumulative returns by factor quantiles in the period range
    defined by -periods_before to periods_after

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    returns : pd.DataFrame
        A wide form Pandas DataFrame indexed by date with assets in the
        columns. Returns data should span the factor analysis time period
        plus/minus an additional buffer window corresponding to periods_after/
        periods_before parameters.
    periods_before : int, optional
        How many periods before factor to plot
    periods_after  : int, optional
        How many periods after factor to plot
    demeaned : bool, optional
        Compute demeaned mean returns (long short portfolio)
    group_adjust : bool
        Returns demeaning will occur on the group level (group
        neutral portfolio)
    by_group : bool
        If True, compute cumulative returns separately for each group

    Returns
    -------
    cumulative returns and std deviation : pd.DataFrame
        A MultiIndex DataFrame indexed by quantile (level 0) and mean/std
        (level 1) and the values on the columns in range from
        -periods_before to periods_after
        If by_group=True the index will have an additional 'group' level
        ::
            ---------------------------------------------------
                        |       | -2  | -1  |  0  |  1  | ...
            ---------------------------------------------------
              quantile  |       |     |     |     |     |
            ---------------------------------------------------
                        | mean  |  x  |  x  |  x  |  x  |
                 1      ---------------------------------------
                        | std   |  x  |  x  |  x  |  x  |
            ---------------------------------------------------
                        | mean  |  x  |  x  |  x  |  x  |
                 2      ---------------------------------------
                        | std   |  x  |  x  |  x  |  x  |
            ---------------------------------------------------
                ...     |                 ...
            ---------------------------------------------------
    """

    def cumulative_return_around_event(q_fact, demean_by):
        return common_start_returns(
            q_fact,
            returns,
            periods_before,
            periods_after,
            cumulative=True,
            mean_by_date=True,
            demean_by=demean_by,
        )

    def average_cumulative_return(q_fact, demean_by):
        q_returns = cumulative_return_around_event(q_fact, demean_by)
        q_returns.replace([np.inf, -np.inf], np.nan, inplace=True)

        return pd.DataFrame({'mean': q_returns.mean(skipna=True, axis=1),
                             'std': q_returns.std(skipna=True, axis=1)}).T

    if by_group:
        #
        # Compute quantile cumulative returns separately for each group
        # Deman those returns accordingly to 'group_adjust' and 'demeaned'
        #
        returns_bygroup = []

        for group, g_data in factor_data.groupby('group'):
            g_fq = g_data['factor_quantile']
            if group_adjust:
                demean_by = g_fq  # demeans at group level
            elif demeaned:
                demean_by = factor_data['factor_quantile']  # demean by all
            else:
                demean_by = None
            #
            # Align cumulative return from different dates to the same index
            # then compute mean and std
            #
            avgcumret = g_fq.groupby(g_fq).apply(average_cumulative_return,
                                                 demean_by)
            if len(avgcumret) == 0:
                continue

            avgcumret['group'] = group
            avgcumret.set_index('group', append=True, inplace=True)
            returns_bygroup.append(avgcumret)

        return pd.concat(returns_bygroup, axis=0)

    else:
        #
        # Compute quantile cumulative returns for the full factor_data
        # Align cumulative return from different dates to the same index
        # then compute mean and std
        # Deman those returns accordingly to 'group_adjust' and 'demeaned'
        #
        if group_adjust:
            all_returns = []
            for group, g_data in factor_data.groupby('group'):
                g_fq = g_data['factor_quantile']
                avgcumret = g_fq.groupby(g_fq).apply(
                    cumulative_return_around_event, g_fq
                )
                all_returns.append(avgcumret)
            q_returns = pd.concat(all_returns, axis=1)
            q_returns = pd.DataFrame({'mean': q_returns.mean(axis=1),
                                      'std': q_returns.std(axis=1)})
            return q_returns.unstack(level=1).stack(level=0)
        elif demeaned:
            fq = factor_data['factor_quantile']
            return fq.groupby(fq).apply(average_cumulative_return, fq)
        else:
            fq = factor_data['factor_quantile']
            return fq.groupby(fq).apply(average_cumulative_return, None)


def factor_cumulative_returns(factor_data,
                              period,
                              long_short=True,
                              group_neutral=False,
                              equal_weight=False,
                              quantiles=None,
                              groups=None):
    """
    Simulate a portfolio using the factor in input and returns the cumulative
    returns of the simulated portfolio

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to,
        and (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    period : string
        'factor_data' column name corresponding to the 'period' returns to be
        used in the computation of porfolio returns
    long_short : bool, optional
        if True then simulates a dollar neutral long-short portfolio
        - see performance.create_pyfolio_input for more details
    group_neutral : bool, optional
        If True then simulates a group neutral portfolio
        - see performance.create_pyfolio_input for more details
    equal_weight : bool, optional
        Control the assets weights:
        - see performance.create_pyfolio_input for more details
    quantiles: sequence[int], optional
        Use only specific quantiles in the computation. By default all
        quantiles are used
    groups: sequence[string], optional
        Use only specific groups in the computation. By default all groups
        are used

    Returns
    -------
    Cumulative returns series : pd.Series
        Example:
            2015-07-16 09:30:00  -0.012143
            2015-07-16 12:30:00   0.012546
            2015-07-17 09:30:00   0.045350
            2015-07-17 12:30:00   0.065897
            2015-07-20 09:30:00   0.030957
    """
    fwd_ret_cols = utils.get_forward_returns_columns(factor_data.columns)

    if period not in fwd_ret_cols:
        raise ValueError("Period '%s' not found" % period)

    todrop = list(fwd_ret_cols)
    todrop.remove(period)
    portfolio_data = factor_data.drop(todrop, axis=1)

    if quantiles is not None:
        portfolio_data = portfolio_data[portfolio_data['factor_quantile'].isin(
            quantiles)]

    if groups is not None:
        portfolio_data = portfolio_data[portfolio_data['group'].isin(groups)]

    returns = \
        factor_returns(portfolio_data, long_short, group_neutral, equal_weight)

    return cumulative_returns(returns[period], period)


def factor_positions(factor_data,
                     period,
                     long_short=True,
                     group_neutral=False,
                     equal_weight=False,
                     quantiles=None,
                     groups=None):
    """
    Simulate a portfolio using the factor in input and returns the assets
    positions as percentage of the total portfolio.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to,
        and (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    period : string
        'factor_data' column name corresponding to the 'period' returns to be
        used in the computation of porfolio returns
    long_short : bool, optional
        if True then simulates a dollar neutral long-short portfolio
        - see performance.create_pyfolio_input for more details
    group_neutral : bool, optional
        If True then simulates a group neutral portfolio
        - see performance.create_pyfolio_input for more details
    equal_weight : bool, optional
        Control the assets weights:
        - see performance.create_pyfolio_input for more details.
    quantiles: sequence[int], optional
        Use only specific quantiles in the computation. By default all
        quantiles are used
    groups: sequence[string], optional
        Use only specific groups in the computation. By default all groups
        are used

    Returns
    -------
    assets positions : pd.DataFrame
        Assets positions series, datetime on index, assets on columns.
        Example:
            index                 'AAPL'         'MSFT'          cash
            2004-01-09 10:30:00   13939.3800     -14012.9930     711.5585
            2004-01-09 15:30:00       0.00       -16012.9930     411.5585
            2004-01-12 10:30:00   14492.6300     -14624.8700       0.0
            2004-01-12 15:30:00   14874.5400     -15841.2500       0.0
            2004-01-13 10:30:00   -13853.2800    13653.6400      -43.6375
    """
    fwd_ret_cols = utils.get_forward_returns_columns(factor_data.columns)

    if period not in fwd_ret_cols:
        raise ValueError("Period '%s' not found" % period)

    todrop = list(fwd_ret_cols)
    todrop.remove(period)
    portfolio_data = factor_data.drop(todrop, axis=1)

    if quantiles is not None:
        portfolio_data = portfolio_data[portfolio_data['factor_quantile'].isin(
            quantiles)]

    if groups is not None:
        portfolio_data = portfolio_data[portfolio_data['group'].isin(groups)]

    weights = \
        factor_weights(portfolio_data, long_short, group_neutral, equal_weight)

    return positions(weights, period)


def create_pyfolio_input(factor_data,
                         period,
                         capital=None,
                         long_short=True,
                         group_neutral=False,
                         equal_weight=False,
                         quantiles=None,
                         groups=None,
                         benchmark_period='1D'):
    """
    Simulate a portfolio using the input factor and returns the portfolio
    performance data properly formatted for Pyfolio analysis.

    For more details on how this portfolio is built see:
    - performance.cumulative_returns (how the portfolio returns are computed)
    - performance.factor_weights (how assets weights are computed)

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to,
        and (optionally) the group the asset belongs to.
        - See full explanation in utils.get_clean_factor_and_forward_returns
    period : string
        'factor_data' column name corresponding to the 'period' returns to be
        used in the computation of porfolio returns
    capital : float, optional
        If set, then compute 'positions' in dollar amount instead of percentage
    long_short : bool, optional
        if True enforce a dollar neutral long-short portfolio: asset weights
        will be computed by demeaning factor values and dividing by the sum of
        their absolute value (achieving gross leverage of 1) which will cause
        the portfolio to hold both long and short positions and the total
        weights of both long and short positions will be equal.
        If False the portfolio weights will be computed dividing the factor
        values and  by the sum of their absolute value (achieving gross
        leverage of 1). Positive factor values will generate long positions and
        negative factor values will produce short positions so that a factor
        with only posive values will result in a long only portfolio.
    group_neutral : bool, optional
        If True simulates a group neutral portfolio: the portfolio weights
        will be computed so that each group will weigh the same.
        if 'long_short' is enabled the factor values demeaning will occur on
        the group level resulting in a dollar neutral, group neutral,
        long-short portfolio.
        If False group information will not be used in weights computation.
    equal_weight : bool, optional
        if True the assets will be equal-weighted. If long_short is True then
        the factor universe will be split in two equal sized groups with the
        top assets in long positions and bottom assets in short positions.
        if False the assets will be factor-weighed, see 'long_short' argument
    quantiles: sequence[int], optional
        Use only specific quantiles in the computation. By default all
        quantiles are used
    groups: sequence[string], optional
        Use only specific groups in the computation. By default all groups
        are used
    benchmark_period : string, optional
        By default benchmark returns are computed as the factor universe mean
        daily returns but 'benchmark_period' allows to choose a 'factor_data'
        column corresponding to the returns to be used in the computation of
        benchmark returns. More generally benchmark returns are computed as the
        factor universe returns traded at 'benchmark_period' frequency, equal
        weighting and long only


    Returns
    -------
     returns : pd.Series
        Daily returns of the strategy, noncumulative.
         - Time series with decimal returns.
         - Example:
            2015-07-16    -0.012143
            2015-07-17    0.045350
            2015-07-20    0.030957
            2015-07-21    0.004902

     positions : pd.DataFrame
        Time series of dollar amount (or percentage when 'capital' is not
        provided) invested in each position and cash.
         - Days where stocks are not held can be represented by 0.
         - Non-working capital is labelled 'cash'
         - Example:
            index         'AAPL'         'MSFT'          cash
            2004-01-09    13939.3800     -14012.9930     711.5585
            2004-01-12    14492.6300     -14624.8700     27.1821
            2004-01-13    -13853.2800    13653.6400      -43.6375


     benchmark : pd.Series
        Benchmark returns computed as the factor universe mean daily returns.

    """

    #
    # Build returns:
    # we don't know the frequency at which the factor returns are computed but
    # pyfolio wants daily returns. So we compute the cumulative returns of the
    # factor, then resample it at 1 day frequency and finally compute daily
    # returns
    #
    cumrets = factor_cumulative_returns(factor_data,
                                        period,
                                        long_short,
                                        group_neutral,
                                        equal_weight,
                                        quantiles,
                                        groups)
    cumrets = cumrets.resample('1D').last().fillna(method='ffill')
    returns = cumrets.pct_change().fillna(0)

    #
    # Build positions. As pyfolio asks for daily position we have to resample
    # the positions returned by 'factor_positions' at 1 day frequency and
    # recompute the weights so that the sum of daily weights is 1.0
    #
    positions = factor_positions(factor_data,
                                 period,
                                 long_short,
                                 group_neutral,
                                 equal_weight,
                                 quantiles,
                                 groups)
    positions = positions.resample('1D').sum().fillna(method='ffill')
    positions = positions.div(positions.abs().sum(axis=1), axis=0).fillna(0)
    positions['cash'] = 1. - positions.sum(axis=1)

    # transform percentage positions to dollar positions
    if capital is not None:
        positions = positions.mul(
            cumrets.reindex(positions.index) * capital, axis=0)

    #
    #
    #
    # Build benchmark returns as the factor universe mean returns traded at
    # 'benchmark_period' frequency
    #
    fwd_ret_cols = utils.get_forward_returns_columns(factor_data.columns)
    if benchmark_period in fwd_ret_cols:
        benchmark_data = factor_data.copy()
        # make sure no negative positions
        benchmark_data['factor'] = benchmark_data['factor'].abs()
        benchmark_rets = factor_cumulative_returns(benchmark_data,
                                                   benchmark_period,
                                                   long_short=False,
                                                   group_neutral=False,
                                                   equal_weight=True)
        benchmark_rets = benchmark_rets.resample(
            '1D').last().fillna(method='ffill')
        benchmark_rets = benchmark_rets.pct_change().fillna(0)
        benchmark_rets.name = 'benchmark'
    else:
        benchmark_rets = None

    return returns, positions, benchmark_rets



import numpy as np
import pandas as pd
from pandas import Series,DataFrame
from .datatools import DataHandle
import statsmodels.api as sm
import scipy.stats as scist
from math import log
from joblib import Parallel,delayed


class StatReturns:
    """
    Input the returns series with freq-D, and get the statistics about the returns. If you input the benchmark returns series, you can get the alpha,beta and so on.

    Params: 
    - returns:DataFrame. Give the returns dataframe with day-freq
    - benchmark:Series. Give the benchmark asset returns series with day-freq. If you don't provide the data, you will not get the correlative analysis results.
    - start_date=None. The start date you analysis, if you do not provide, start date will be the returns data's start.
    - end_date=None. The end date you analysis, if you do not provide, end date will be the returns data's end.

    """
    def __init__(self,returns:DataFrame,benchmark:Series=None,start_date=None,end_date=None):
        if start_date is None:
            self._start_date = returns.index[0]
        else:
            self._start_date = start_date
        if end_date is None:
            self._end_date = returns.index[-1]
        else:
            self._end_date = end_date
        
        self._returns = returns.loc[start_date:end_date].copy()
        self._pnl = (1+self._returns).cumprod()
        if benchmark is not None:
            self._benchmark = benchmark.loc[start_date:end_date]
        else:
            self._benchmark = None


    def stat_frame(self,market:bool=False,window=90,q=0.05,rf=0):
        """
        Params:
        - market:bool=False. If market bool is True, you will get the market-correlative analysis, but ensure you provide the benchmark data.
        - window=90. Natual day number when you calculate the max drawdown, and it mutiply 21/30 as the trade day number.
        - q=0.05. VaR quantile. 
        - rf=0. Annualized risk free rate, this will be used when you calculate the sharpe ratio and linear regression.

        Results:Dataframe. With asset index and analysis project columns.

        Attentions:
        - 252 trading days a year.

        """
        means = self._returns.mean()*252
        stds = self._returns.std()*np.sqrt(252)
        sharpe_ratio = (means-rf)/stds
        max_drawdown = (1-self._pnl/(self._pnl.rolling(window=int(window*21/30)).max())).max() #每30天约为21个交易日，window为自然日
        VaR = self._returns.quantile(q)

        if not market:
            result = pd.concat([means,stds,sharpe_ratio,max_drawdown,VaR],axis=1)
            result.columns = ['mean','std','SR',f'mdd_{window}D',f'VaR_{q}']
            return result
        
        if self._benchmark is None:
            raise ValueError('You have not provide the benchmark returns.')
        
        alpha = []
        beta = []
        residual_std = []
        X,y = DataHandle.handle_nan_value_X_y(self._returns,self._benchmark,method='mfill')
        for col in self._returns.columns:
            params = np.polyfit(X[col],y,deg=1)
            be,al = params
            res = np.std(y-np.polyval(params,X[col]))
            beta.append(be)
            alpha.append(al*252)
            residual_std.append(res)
        beta = Series(beta,index = X.columns)
        alpha = Series(alpha, index = X.columns)-rf
        residual_std = Series(residual_std,index=X.columns)
        TR = (means-rf)/beta
        IR = alpha/residual_std/np.sqrt(252)
        result = pd.concat([means,stds,sharpe_ratio,max_drawdown,VaR,beta,alpha,TR,IR],axis=1)
        result.columns = ['mean','std','SR',f'mdd_{window}D',f'VaR_{q}','beta','alpha','TR','IR']
        result.index.name = 'asset'
        return result


def stat_returns(returns,benchmark=None,start_date=None,end_date=None,market=False,window=90,q=0.05,rf=0):
    stat = StatReturns(returns=returns,benchmark=benchmark,start_date=start_date,end_date=end_date)
    frame = stat.stat_frame(market=market,window=window,q=q,rf=rf)
    return frame

def stat_returns_by_period(returns,skip=100,benchmark=None,start_date=None,end_date=None,market=False,window=90,q=0.05,rf=0):
    result = []
    for k in range(0,len(returns.index),skip):
        df = returns.iloc[k:k+skip].copy()
        frame = stat_returns(df,benchmark,start_date=None,end_date=None,market=market,window=window,q=q,rf=rf)
        frame['start_date'] = df.index[0]
        frame['end_date'] = df.index[-1]
        result.append(frame)
    result = pd.concat(result)
    result.index.name = 'asset'
    result = result.reset_index().set_index(['end_date','asset']).sort_index()
    return result
        

class StatTtest:
    def __init__(self,X:Series,y:Series,method):
        data = pd.concat([X,y],axis=1).dropna()
        data.columns = ['X','y']
        self.data = data
        self.X = data['X']
        self.y = data['y']
        self.degree = len(self.data.index)
        self.method = method
    
    def reg_coef(self):
        if len(self.X) != len(self.y):
            raise ValueError("X,y should have the same dimension")
        if self.method == 'OLS':
            X = sm.add_constant(self.data['X'].values)
            y = self.data['y'].values
            model = sm.OLS(y,X).fit()
            beta = model.params[1]
            sr = model.bse[1]
            return (beta, beta/sr)
        if self.method == 'RLM':
            X = sm.add_constant(self.data['X'].values)
            y = self.data['y'].values
            model = sm.RLM(y,X).fit()
            beta = model.params[1]
            sr = model.bse[1]
            return (beta, beta/sr)
        if self.method == 'QR':
            X = sm.add_constant(self.data['X'].values)
            y = self.data['y'].values
            model = sm.QuantReg(y,X).fit(max_iter=3000)
            beta = model.params[1]
            sr = model.bse[1]
            return (beta, beta/sr)
        raise ValueError("method should be OLS, RLM or QR")
    
    def t_test(self,alpha=0.05):
        """返回系数是否大于0，是否显著大于0和小于0的bool值列表"""
        beta, t_value = self.reg_coef()
        bool_lst = [beta>0]
        if t_value > scist.t.ppf(1-alpha,df=self.degree-2):
            bool_lst.append(True)
        else:
            bool_lst.append(False)
        if t_value < scist.t.ppf(alpha,df=self.degree-2):
            bool_lst.append(True)
        else:
            bool_lst.append(False)
        return bool_lst
    

def order_array(value:np.ndarray):
    v = np.array(value)
    if len(v.shape) > 1:
        raise ValueError
    l = len(v)
    order_lst = [0 for i in range(l)]
    argsort = np.argsort(v)
    for i in range(l):
        order_lst[argsort[i]] = i 
    return np.array(order_lst)

class TabelRegression:
    """提供一个DataFrame, columns[0]=y, columns[1:]=X, index(level=0)=date, index(level=1)=asset"""
    
    def __init__(self,df:pd.DataFrame, n_jobs=12) -> None:
        self.df = df.dropna()
        self.y = self.df[df.columns[0]]
        self.X = self.df[df.columns[1:]]
        self.dates = self.df.index.get_level_values(0).unique()
        self.n_jobs = n_jobs
    
    def corr(self):
        return self.df.groupby(level=0).corr(method='spearman').groupby(level=1).mean()
    
    def mean(self, df):
        return df.groupby(level=0).mean()
    
    def std(self, df):
        return df.groupby(level=0).std()
    
    def linear_regression(self):
        
        def work_func(date,X,y):
            X = sm.add_constant(X)
            model = sm.OLS(y,X).fit()
            resid = model.resid
            params = model.params
            return (resid,params)

        
        results = Parallel(n_jobs=self.n_jobs)(delayed(work_func)(date,self.X.loc[date].values,self.y.loc[date].values) for date in self.dates)
        params = [np.array(p) for _,p in results if p is not None]
        resids = [np.array(r) for r,_ in results if r is not None]
        #### 注意如果你的模型某些日期有常数列，会导致参数确实一位，parmas无法stack，谨慎检查
        # params = pd.DataFrame(np.vstack(params),index=self.dates)
        resids = pd.DataFrame(np.hstack(resids),index=self.df.index)
        return (params,resids)
    
    def order_regression(self):
        def work_func(date,X,y):
            X = X.T
            result = []
            for i in range(len(X)):
                result.append(order_array(X[i]))
            X = np.vstack(result).T
            X = X/len(X)
            y = order_array(y)/len(y)
            X = sm.add_constant(X)
            model = sm.OLS(y,X).fit()
            resid = model.resid
            params = model.params
            return (resid,params)
        
        results = Parallel(n_jobs=self.n_jobs)(delayed(work_func)(date,self.X.loc[date].values,self.y.loc[date].values) for date in self.dates)
        params = [np.array(p) for _,p in results if p is not None]
        resids = [np.array(r) for r,_ in results if r is not None]
        #### 注意如果你的模型某些日期有常数列，会导致参数确实一位，parmas无法stack，谨慎检查
        # params = pd.DataFrame(np.vstack(params),index=self.dates)
        resids = pd.DataFrame(np.hstack(resids),index=self.df.index)
        return (params,resids)
            
            
  


from .cta_alpha_function import *
import pandas as pd
class AlphaLC(object):
    def __init__(self, df_data):

        self.open = df_data['open']
        self.high = df_data['high']
        self.low = df_data['low']
        self.close = df_data['close']
        self.volume = df_data['volume']
        self.pct_chg = (self.close.ffill()/(self.close.ffill().shift(1)) - 1)
        self.amt = df_data['amount']
        self.swing = (self.high.values - self.low.values)/ (self.close.shift(1))
        self.turn = df_data['turnover']
        self.vwap = (df_data['amount']*1000)/(df_data['volume']*100+1) 
        # 因为成交
        self.volume_ratio = (self.volume/self.volume.shift(1) - 1).replace({np.inf: np.nan, -np.inf: np.nan}) #加0.01是为了防止分母出现0,用1去填充第一个值
        # Alpha#1	 (rank(Ts_ArgMax(SignedPower(((returns < 0) ? stddev(returns, 20) : close), 2.), 5)) -0.5)

        # self.open.columns = ['data']
        # self.close.columns = ['data']
        # self.high.columns = ['data']
        # self.low.columns = ['data']
        # self.volume.columns = ['data']
        # self.amt.columns = ['data']
        # self.turn.columns = ['data']
        # self.vwap.columns = ['data']
        # self.volume_ratio.columns = ['data']
        
    def basic_in1(self):
        return self.open
        
    def basic_in2(self):
        return self.close
        
    def basic_in3(self):
        return self.high
        
    def basic_in4(self):
        return self.low
        
    def basic_in5(self):
        return self.volume
        
    def basic_in6(self):
        return self.amt
        
    def basic_in7(self):
        return self.turn
        
    def basic_in8(self):
        return self.swing
        
    def basic_in9(self):
        return self.pct_chg
        
    def basic_in10(self):
        return self.vwap
        
    def basic_in12(self):
        return self.volume_ratio
        
    def basic_in13(self):
        return self.close.shift(1)
        
    def basic_in14(self):
        return self.volume.shift(1)
        
    def basic_in15(self):
        return self.swing.shift(1)
        
    def basic_in16(self):
        return self.pct_chg.shift(1)
        
    def mean_in1(self):
        return Ts_Single_Mean(self.volume,3)
        
    def mean_in2(self):
        return Ts_Single_Mean(self.volume,5)
        
    def mean_in3(self):
        return Ts_Single_Mean(self.volume,10)
        
    def mean_in4(self):
        return Ts_Single_Mean(self.volume,30)
        
    def mean_in5(self):
        return Ts_Single_Mean(self.volume,60)
        
    def mean_in6(self):
        return Ts_Single_Mean(self.pct_chg,3)
        
    def mean_in7(self):
        return Ts_Single_Mean(self.pct_chg,5)
        
    def mean_in8(self):
        return Ts_Single_Mean(self.pct_chg,10)
        
    def mean_in9(self):
        return Ts_Single_Mean(self.pct_chg,30)
        
    def mean_in10(self):
        return Ts_Single_Mean(self.pct_chg,60)
        
    def mean_in11(self):
        return Ts_Single_Mean(self.amt*Self_Sign(self.pct_chg),3)
        
    def mean_in12(self):
        return Ts_Single_Mean(self.amt*Self_Sign(self.pct_chg),5)
        
    def mean_in13(self):
        return Ts_Single_Mean(self.amt*Self_Sign(self.pct_chg),10)
        
    def mean_in14(self):
        return Ts_Single_Mean(self.amt*Self_Sign(self.pct_chg),30)
        
    def mean_in15(self):
        return Ts_Single_Mean(self.amt*Self_Sign(self.pct_chg),60)
        
    def mean_in16(self):
        return Ts_Single_Mean(self.turn,3)
        
    def mean_in17(self):
        return Ts_Single_Mean(self.turn,5)
        
    def mean_in18(self):
        return Ts_Single_Mean(self.turn,10)
        
    def mean_in19(self):
        return Ts_Single_Mean(self.turn,30)
        
    def mean_in20(self):
        return Ts_Single_Mean(self.turn,60)
        
    def mean_in21(self):
        return Ts_Single_Mean(self.volume_ratio,3)
        
    def mean_in22(self):
        return Ts_Single_Mean(self.volume_ratio,5)
        
    def mean_in23(self):
        return Ts_Single_Mean(self.volume_ratio,10)
        
    def mean_in24(self):
        return Ts_Single_Mean(self.volume_ratio,30)
        
    def mean_in25(self):
        return Ts_Single_Mean(self.volume_ratio,60)
        
    def delta_in1(self):
        return Ts_Single_Delta(self.turn,1)
        
    def delta_in2(self):
        return Ts_Single_Delta(self.volume,1)
        
    def delta_in3(self):
        return Ts_Single_Delta(self.pct_chg,1)
        
    def delta_in5(self):
        return Ts_Single_Delta(self.amt*Self_Sign(self.pct_chg),1)
        
    def delta_in6(self):
        return Ts_Single_Delta(self.open,1)
        
    def delta_in7(self):
        return Ts_Single_Delta(self.close,1)
        
    def delta_in8(self):
        return Ts_Single_Delta(self.high,1)
        
    def delta_in9(self):
        return Ts_Single_Delta(self.low,1)
        
    def delta_in10(self):
        return Ts_Single_Delta(self.vwap,1)
        
    def delta_in11(self):
        return Ts_Single_Rate(self.vwap,1)
        
    def delta_in13(self):
        return Ts_Single_Rate(self.turn,1)
        
    def delta_in14(self):
        return Ts_Single_Rate(self.volume,1)
        
    def delta_in15(self):
        return Ts_Single_Rate(self.pct_chg,1)
        
    def delta_in18(self):
        return Ts_Single_Rate(self.amt*Self_Sign(self.pct_chg),1)
        
    def delta_in19(self):
        return Ts_Single_Rate(self.open,1)
        
    def delta_in20(self):
        return Ts_Single_Rate(self.close,1)
        
    def delta_in21(self):
        return Ts_Single_Rate(self.high,1)
        
    def delta_in22(self):
        return Ts_Single_Rate(self.low,1)
        
    def delta_in23(self):
        return Ts_Single_Delta(self.volume_ratio,1)
        
    def delta_in24(self):
        return Ts_Single_Rate(self.volume_ratio,1)
        
    def mometum_in1(self):
        return Cs_Single_UpZeroLimit(Ts_Single_Sum(self.pct_chg,3))
        
    def mometum_in2(self):
        return Cs_Single_UpZeroLimit(Ts_Single_Sum(self.pct_chg,5))
        
    def mometum_in3(self):
        return Cs_Single_UpZeroLimit(Ts_Single_Sum(self.pct_chg,10))
        
    def mometum_in4(self):
        return Cs_Single_UpZeroLimit(Ts_Single_Sum(self.pct_chg,30))
        
    def mometum_in5(self):
        return Cs_Single_UpZeroLimit(Ts_Single_Sum(self.pct_chg,60))
        
    def mometum_in6(self):
        return Ts_Shape_Compare(self.open,self.close,3)/3
        
    def mometum_in7(self):
        return Ts_Shape_Compare(self.open,self.close,5)/5
        
    def mometum_in8(self):
        return Ts_Shape_Compare(self.open,self.close,10)/10
        
    def mometum_in9(self):
        return Ts_Shape_Compare(self.open,self.close,30)/30
        
    def mometum_in10(self):
        return Ts_Shape_Compare(self.open,self.close,60)/60
        
    def mometum_in11(self):
        return Ts_Increase_Times_Discontinuous(self.close,3)/3
        
    def mometum_in12(self):
        return Ts_Increase_Times_Discontinuous(self.close,5)/5
        
    def mometum_in13(self):
        return Ts_Increase_Times_Discontinuous(self.close,10)/10
        
    def mometum_in14(self):
        return Ts_Increase_Times_Discontinuous(self.close,30)/30
        
    def mometum_in15(self):
        return Ts_Increase_Times_Discontinuous(self.close,60)/60
        
    def mometum_in16(self):
        return Ts_Increase_Times_Continuous(self.close,3)/3
        
    def mometum_in17(self):
        return Ts_Increase_Times_Continuous(self.close,5)/5
        
    def mometum_in18(self):
        return Ts_Increase_Times_Continuous(self.close,10)/10
        
    def mometum_in19(self):
        return Ts_Increase_Times_Continuous(self.close,30)/30
        
    def mometum_in20(self):
        return Ts_Increase_Times_Continuous(self.close,60)/60
        
    def mometum_in21(self):
        return Ts_Single_RSI(self.pct_chg,3)
        
    def mometum_in22(self):
        return Ts_Single_RSI(self.pct_chg,5)
        
    def mometum_in23(self):
        return Ts_Single_RSI(self.pct_chg,10)
        
    def mometum_in24(self):
        return Ts_Single_RSI(self.pct_chg,30)
        
    def mometum_in25(self):
        return Ts_Single_RSI(self.pct_chg,60)
        
    def mometum_in26(self):
        return Cs_Double_Multiply(Ts_Single_Rate(self.close,3),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),3),1.5))
        
    def mometum_in27(self):
        return Cs_Double_Multiply(Ts_Single_Rate(self.close,5),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),5),1.5))
        
    def mometum_in28(self):
        return Cs_Double_Multiply(Ts_Single_Rate(self.close,7),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),7),1.5))
        
    def mometum_in29(self):
        return Cs_Double_Multiply(Ts_Single_Rate(self.close,14),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),14),1.5))
        
    def mometum_in30(self):
        return Cs_Double_Multiply(Ts_Single_Rate(self.close,28),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),28),1.5))
        
    def reversion_in1(self):
        return Self_Negative(Cs_Single_DownZeroLimit(Ts_Single_Sum(self.pct_chg,3)))
        
    def reversion_in2(self):
        return Self_Negative(Cs_Single_DownZeroLimit(Ts_Single_Sum(self.pct_chg,5)))
        
    def reversion_in3(self):
        return Self_Negative(Cs_Single_DownZeroLimit(Ts_Single_Sum(self.pct_chg,10)))
        
    def reversion_in4(self):
        return Self_Negative(Cs_Single_DownZeroLimit(Ts_Single_Sum(self.pct_chg,30)))
        
    def reversion_in5(self):
        return Self_Negative(Cs_Single_DownZeroLimit(Ts_Single_Sum(self.pct_chg,60)))
        
    def reversion_in6(self):
        return 1-Ts_Shape_Compare(self.open,self.close,3)/3
        
    def reversion_in7(self):
        return 1-Ts_Shape_Compare(self.open,self.close,5)/5
        
    def reversion_in8(self):
        return 1-Ts_Shape_Compare(self.open,self.close,10)/10
        
    def reversion_in9(self):
        return 1-Ts_Shape_Compare(self.open,self.close,22)/22
        
    def reversion_in10(self):
        return 1-Ts_Shape_Compare(self.open,self.close,60)/60
        
    def reversion_in11(self):
        return Ts_Decrease_Times_Discontinuous(self.close,3)/3
        
    def reversion_in12(self):
        return Ts_Decrease_Times_Discontinuous(self.close,5)/5
        
    def reversion_in13(self):
        return Ts_Decrease_Times_Discontinuous(self.close,10)/10
        
    def reversion_in14(self):
        return Ts_Decrease_Times_Discontinuous(self.close,30)/30
        
    def reversion_in15(self):
        return Ts_Decrease_Times_Discontinuous(self.close,60)/60
        
    def reversion_in16(self):
        return Ts_Decrease_Times_Continuous(self.close,3)/3
        
    def reversion_in17(self):
        return Ts_Decrease_Times_Continuous(self.close,5)/5
        
    def reversion_in18(self):
        return Ts_Decrease_Times_Continuous(self.close,10)/10
        
    def reversion_in19(self):
        return Ts_Decrease_Times_Continuous(self.close,30)/30
        
    def reversion_in20(self):
        return Ts_Decrease_Times_Continuous(self.close,60)/60
        
    def reversion_in21(self):
        return Ts_Single_RSI(-self.pct_chg,3)
        
    def reversion_in22(self):
        return Ts_Single_RSI(-self.pct_chg,5)
        
    def reversion_in23(self):
        return Ts_Single_RSI(-self.pct_chg,10)
        
    def reversion_in24(self):
        return Ts_Single_RSI(-self.pct_chg,22)
        
    def reversion_in25(self):
        return Ts_Single_RSI(-self.pct_chg,60)
        
    def reversion_in26(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,3),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),3),1.5))
        
    def reversion_in27(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,5),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),5),1.5))
        
    def reversion_in28(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,7),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),7),1.5))
        
    def reversion_in29(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,14),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),14),1.5))
        
    def reversion_in30(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,28),Self_Power(Ts_Single_Linear_Decay(Ts_Single_Divide(self.volume,1),28),1.5))
        
    def volume_in1(self):
        return Ts_Single_AccRate(self.volume,3)
        
    def volume_in2(self):
        return Ts_Single_AccRate(self.volume,5)
        
    def volume_in3(self):
        return Ts_Single_AccRate(self.volume,10)
        
    def volume_in4(self):
        return Ts_Single_AccRate(self.volume,20)
        
    def volume_in5(self):
        return Ts_Single_AccRate(self.volume,30)
        
    def volume_in6(self):
        return Ts_Single_Sum(Self_Sign(self.pct_chg)*self.volume,3)
        
    def volume_in7(self):
        return Ts_Single_Sum(Self_Sign(self.pct_chg)*self.volume,5)
        
    def volume_in8(self):
        return Ts_Single_Sum(Self_Sign(self.pct_chg)*self.volume,10)
        
    def volume_in9(self):
        return Cs_Double_Divide(Ts_Single_Sum(Self_Sign(self.pct_chg)*self.amt,3),self.amt)
        
    def volume_in10(self):
        return Cs_Double_Divide(Ts_Single_Sum(Self_Sign(self.pct_chg)*self.amt,5),self.amt)
        
    def volume_in11(self):
        return Cs_Double_Divide(Ts_Single_Sum(Self_Sign(self.pct_chg)*self.amt,10),self.amt)
        
    def volatility_in1(self):
        return Ts_Single_Stddev(self.pct_chg,3)
        
    def volatility_in2(self):
        return Ts_Single_Stddev(self.pct_chg,5)
        
    def volatility_in3(self):
        return Ts_Single_Stddev(self.pct_chg,10)
        
    def volatility_in4(self):
        return Ts_Single_Stddev(self.pct_chg,30)
        
    def volatility_in5(self):
        return Ts_Single_Stddev(self.pct_chg,60)
        
    def volatility_in6(self):
        return Ts_Single_Stddev(self.swing,3)
        
    def volatility_in7(self):
        return Ts_Single_Stddev(self.swing,5)
        
    def volatility_in8(self):
        return Ts_Single_Stddev(self.swing,10)
        
    def volatility_in9(self):
        return Ts_Single_Stddev(self.swing,30)
        
    def volatility_in10(self):
        return Ts_Single_Stddev(self.swing,60)
        
    def volatility_in11(self):
        return Ts_Single_Stddev(self.turn,3)
        
    def volatility_in12(self):
        return Ts_Single_Stddev(self.turn,5)
        
    def volatility_in13(self):
        return Ts_Single_Stddev(self.turn,10)
        
    def volatility_in14(self):
        return Ts_Single_Stddev(self.turn,30)
        
    def volatility_in15(self):
        return Ts_Single_Stddev(self.turn,60)
        
    def volatility_in16(self):
        return Ts_Single_Max_Divide_Min(self.close,5)-1
        
    def volatility_in17(self):
        return Ts_Single_Max_Divide_Min(self.close,10)-1
        
    def volatility_in18(self):
        return Ts_Single_Max_Divide_Min(self.close,30)-1
        
    def volatility_in19(self):
        return Ts_Single_Max_Divide_Min(self.close,60)-1
        
    def volatility_in20(self):
        return Ts_Single_Max_Divide_Min(self.volume,5)-1
        
    def volatility_in21(self):
        return Ts_Single_Max_Divide_Min(self.volume,10)-1
        
    def volatility_in22(self):
        return Ts_Single_Max_Divide_Min(self.volume,30)-1
        
    def volatility_in23(self):
        return Ts_Single_Max_Divide_Min(self.volume,60)-1
        
    def skewness_in1(self):
        return Ts_Single_Skewness(self.volume,5)
        
    def skewness_in2(self):
        return Ts_Single_Skewness(self.volume,10)
        
    def skewness_in3(self):
        return Ts_Single_Skewness(self.volume,30)
        
    def skewness_in4(self):
        return Ts_Single_Skewness(self.volume,60)
        
    def skewness_in5(self):
        return Ts_Single_Skewness(self.turn,5)
        
    def skewness_in6(self):
        return Ts_Single_Skewness(self.turn,10)
        
    def skewness_in7(self):
        return Ts_Single_Skewness(self.turn,20)
        
    def skewness_in8(self):
        return Ts_Single_Skewness(self.turn,60)
        
    def skewness_in9(self):
        return Ts_Single_Skewness(self.pct_chg,5)
        
    def skewness_in10(self):
        return Ts_Single_Skewness(self.pct_chg,10)
        
    def skewness_in11(self):
        return Ts_Single_Skewness(self.pct_chg,30)
        
    def skewness_in12(self):
        return Ts_Single_Skewness(self.pct_chg,60)
        
    def skewness_in13(self):
        return Ts_Single_Skewness(self.open,5)
        
    def skewness_in14(self):
        return Ts_Single_Skewness(self.open,10)
        
    def skewness_in15(self):
        return Ts_Single_Skewness(self.open,30)
        
    def skewness_in16(self):
        return Ts_Single_Skewness(self.open,60)
        
    def skewness_in21(self):
        return Ts_Double_Skewness(self.turn,self.swing,5)
        
    def skewness_in22(self):
        return Ts_Double_Skewness(self.turn,self.swing,10)
        
    def skewness_in23(self):
        return Ts_Double_Skewness(self.pct_chg,self.turn,5)
        
    def skewness_in24(self):
        return Ts_Double_Skewness(self.pct_chg,self.turn,10)
        
    def skewness_in25(self):
        return Ts_Double_Skewness(self.pct_chg,self.swing,5)
        
    def skewness_in26(self):
        return Ts_Double_Skewness(self.pct_chg,self.swing,10)
        
    def skewness_in27(self):
        return Ts_Double_Skewness(self.open,self.close,5)
        
    def skewness_in28(self):
        return Ts_Double_Skewness(self.open,self.close,10)
        
    def skewness_in29(self):
        return Ts_Double_Skewness(self.high,self.low,5)
        
    def skewness_in30(self):
        return Ts_Double_Skewness(self.high,self.low,10)
        
    def kurtosis_in1(self):
        return Ts_Single_Kurtosis(self.volume,5)
        
    def kurtosis_in2(self):
        return Ts_Single_Kurtosis(self.volume,10)
        
    def kurtosis_in3(self):
        return Ts_Single_Kurtosis(self.volume,30)
        
    def kurtosis_in4(self):
        return Ts_Single_Kurtosis(self.volume,60)
        
    def kurtosis_in5(self):
        return Ts_Single_Kurtosis(self.turn,5)
        
    def kurtosis_in6(self):
        return Ts_Single_Kurtosis(self.turn,10)
        
    def kurtosis_in7(self):
        return Ts_Single_Kurtosis(self.turn,30)
        
    def kurtosis_in8(self):
        return Ts_Single_Kurtosis(self.turn,60)
        
    def kurtosis_in9(self):
        return Ts_Single_Kurtosis(self.pct_chg,5)
        
    def kurtosis_in10(self):
        return Ts_Single_Kurtosis(self.pct_chg,10)
        
    def kurtosis_in11(self):
        return Ts_Single_Kurtosis(self.pct_chg,30)
        
    def kurtosis_in12(self):
        return Ts_Single_Kurtosis(self.pct_chg,60)
        
    def kurtosis_in13(self):
        return Ts_Single_Kurtosis(self.open,5)
        
    def kurtosis_in14(self):
        return Ts_Single_Kurtosis(self.open,10)
        
    def kurtosis_in15(self):
        return Ts_Single_Kurtosis(self.open,30)
        
    def kurtosis_in16(self):
        return Ts_Single_Kurtosis(self.open,60)
        
    def kurtosis_in21(self):
        return Ts_Double_Kurtosis(self.turn,self.swing,5)
        
    def kurtosis_in22(self):
        return Ts_Double_Kurtosis(self.turn,self.swing,10)
        
    def kurtosis_in23(self):
        return Ts_Double_Kurtosis(self.pct_chg,self.turn,5)
        
    def kurtosis_in24(self):
        return Ts_Double_Kurtosis(self.pct_chg,self.turn,10)
        
    def kurtosis_in25(self):
        return Ts_Double_Kurtosis(self.pct_chg,self.swing,5)
        
    def kurtosis_in26(self):
        return Ts_Double_Kurtosis(self.pct_chg,self.swing,10)
        
    def kurtosis_in27(self):
        return Ts_Double_Kurtosis(self.open,self.close,5)
        
    def kurtosis_in28(self):
        return Ts_Double_Kurtosis(self.open,self.close,10)
        
    def kurtosis_in29(self):
        return Ts_Double_Kurtosis(self.high,self.low,5)
        
    def kurtosis_in30(self):
        return Ts_Double_Kurtosis(self.high,self.low,10)
        
    def correlation_in1(self):
        return Cs_Single_Correlation(self.pct_chg,3)
        
    def correlation_in2(self):
        return Cs_Single_Correlation(self.pct_chg,5)
        
    def correlation_in3(self):
        return Cs_Single_Correlation(self.pct_chg,10)
        
    def correlation_in4(self):
        return Cs_Single_Correlation(self.pct_chg,30)
        
    def correlation_in5(self):
        return Cs_Single_Correlation(self.turn,3)
        
    def correlation_in6(self):
        return Cs_Single_Correlation(self.turn,5)
        
    def correlation_in7(self):
        return Cs_Single_Correlation(self.turn,10)
        
    def correlation_in8(self):
        return Cs_Single_Correlation(self.turn,30)
        
    def correlation_in9(self):
        return Cs_Single_Correlation(self.volume,3)
        
    def correlation_in10(self):
        return Cs_Single_Correlation(self.volume,5)
        
    def correlation_in11(self):
        return Cs_Single_Correlation(self.volume,10)
        
    def correlation_in12(self):
        return Cs_Single_Correlation(self.volume,30)
        
    def correlation_in13(self):
        return Ts_Double_Correlation(self.pct_chg,self.turn,3)
        
    def correlation_in14(self):
        return Ts_Double_Correlation(self.pct_chg,self.turn,5)
        
    def correlation_in15(self):
        return Ts_Double_Correlation(self.pct_chg,self.turn,10)
        
    def correlation_in16(self):
        return Ts_Double_Correlation(self.pct_chg,self.turn,30)
        
    def correlation_in17(self):
        return Ts_Double_Correlation(self.pct_chg,self.turn,60)
        
    def correlation_in23(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,5,1),Self_Sign(Ts_Single_Delta(self.close,5)))
        
    def correlation_in24(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,10,1),Self_Sign(Ts_Single_Delta(self.close,10)))
        
    def correlation_in25(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,20,1),Self_Sign(Ts_Single_Delta(self.close,20)))
        
    def correlation_in26(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,60,1),Self_Sign(Ts_Single_Delta(self.close,60)))
        
    def correlation_in27(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,6,3),Self_Sign(Ts_Single_Delta(self.close,3)))
        
    def correlation_in28(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,10,5),Self_Sign(Ts_Single_Delta(self.close,5)))
        
    def correlation_in29(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,20,10),Self_Sign(Ts_Single_Delta(self.close,10)))
        
    def correlation_in30(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.close,60,30),Self_Sign(Ts_Single_Delta(self.close,30)))
        
    def correlation_in31(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,5,1),Self_Sign(Ts_Single_Delta(self.volume,5)))
        
    def correlation_in32(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,10,1),Self_Sign(Ts_Single_Delta(self.volume,10)))
        
    def correlation_in33(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,20,1),Self_Sign(Ts_Single_Delta(self.volume,20)))
        
    def correlation_in34(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,60,1),Self_Sign(Ts_Single_Delta(self.volume,60)))
        
    def correlation_in35(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,6,3),Self_Sign(Ts_Single_Delta(self.volume,3)))
        
    def correlation_in36(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,10,5),Self_Sign(Ts_Single_Delta(self.volume,5)))
        
    def correlation_in37(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,20,10),Self_Sign(Ts_Single_Delta(self.volume,10)))
        
    def correlation_in38(self):
        return Cs_Double_Multiply(Ts_Single_AutoCorrelation(self.volume,60,30),Self_Sign(Ts_Single_Delta(self.volume,30)))
        
    def bias_in6(self):
        return Cs_Double_Divide(Ts_Single_Linear_Decay(self.close,3),self.close)-1
        
    def bias_in7(self):
        return Cs_Double_Divide(Ts_Single_Linear_Decay(self.close,5),self.close)-1
        
    def bias_in8(self):
        return Cs_Double_Divide(Ts_Single_Linear_Decay(self.close,10),self.close)-1
        
    def bias_in9(self):
        return Cs_Double_Divide(Ts_Single_Linear_Decay(self.close,30),self.close)-1
        
    def bias_in10(self):
        return Cs_Double_Divide(Ts_Single_Linear_Decay(self.close,60),self.close)-1
        
    def bias_in11(self):
        return Cs_Double_Divide(Ts_Single_Pressure(self.close,7),self.close)-1
        
    def bias_in12(self):
        return Cs_Double_Divide(Ts_Single_Pressure(self.close,14),self.close)-1
        
    def bias_in13(self):
        return Cs_Double_Divide(Ts_Single_Pressure(self.close,30),self.close)-1
        
    def bias_in14(self):
        return Cs_Double_Divide(Ts_Single_Pressure(self.close,60),self.close)-1
        
    def bias_in15(self):
        return Cs_Double_Divide(self.close,Ts_Single_Support(self.close,7))-1
        
    def bias_in16(self):
        return Cs_Double_Divide(self.close,Ts_Single_Support(self.close,14))-1
        
    def bias_in17(self):
        return Cs_Double_Divide(self.close,Ts_Single_Support(self.close,30))-1
        
    def bias_in18(self):
        return Cs_Double_Divide(self.close,Ts_Single_Support(self.close,60))-1
        
    def bias_in19(self):
        return Cs_Double_Divide(self.close,Ts_Single_Max(self.high,5))-0.5
        
    def bias_in20(self):
        return Cs_Double_Divide(self.close,Ts_Single_Max(self.high,10))-0.5
        
    def bias_in21(self):
        return Cs_Double_Divide(self.close,Ts_Single_Max(self.high,20))-0.5
        
    def bias_in22(self):
        return Cs_Double_Divide(self.close,Ts_Single_Max(self.high,60))-0.5
        
    def bias_in23(self):
        return Cs_Double_Divide(self.close,Ts_Single_Min(self.low,5))-0.5
        
    def bias_in24(self):
        return Cs_Double_Divide(self.close,Ts_Single_Min(self.low,10))-0.5
        
    def bias_in25(self):
        return Cs_Double_Divide(self.close,Ts_Single_Min(self.low,20))-0.5
        
    def bias_in26(self):
        return Cs_Double_Divide(self.close,Ts_Single_Min(self.low,60))-0.5
        
    def bias_in27(self):
        return Cs_Double_Divide(self.close,0.01+Cs_Double_Divide(Ts_Single_Sum(self.amt,3),Ts_Single_Sum(self.volume,3)))-1
        
    def bias_in28(self):
        return Cs_Double_Divide(self.close,0.01+Cs_Double_Divide(Ts_Single_Sum(self.amt,5),Ts_Single_Sum(self.volume,5)))-1
        
    def bias_in29(self):
        return Cs_Double_Divide(self.close,0.01+Cs_Double_Divide(Ts_Single_Sum(self.amt,10),Ts_Single_Sum(self.volume,10)))-1
        
    def bias_in30(self):
        return Cs_Double_Divide(self.close,0.01+Cs_Double_Divide(Ts_Single_Sum(self.amt,20),Ts_Single_Sum(self.volume,20)))-1
        
    def bias_in31(self):
        return Cs_Double_Divide(self.close,0.01+Cs_Double_Divide(Ts_Single_Sum(self.amt,60),Ts_Single_Sum(self.volume,60)))-1
        
    def bias_in32(self):
        return Cs_Double_Divide(self.close,self.vwap)-1
        
    def bias_in33(self):
        return Cs_Double_Divide(self.high,self.vwap)-1
        
    def bias_in34(self):
        return Cs_Double_Divide(self.low,self.vwap)-1
        
    def bias_in35(self):
        return Cs_Double_Divide(self.open,self.vwap)-1
        
    def bias_in36(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,5),Ts_Single_Mean(self.close,10))-1
        
    def bias_in37(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,10),Ts_Single_Mean(self.close,20))-1
        
    def bias_in38(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,20),Ts_Single_Mean(self.close,30))-1
        
    def bias_in39(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,30),Ts_Single_Mean(self.close,60))-1
        
    def bias_in40(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,5),Ts_Single_Mean(self.close,10))-1
        
    def bias_in41(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,10),Ts_Single_Mean(self.close,20))-1
        
    def bias_in42(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,20),Ts_Single_Mean(self.close,30))-1
        
    def bias_in43(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.close,30),Ts_Single_Mean(self.close,60))-1
        
    def bias_in44(self):
        return Cs_Double_Divide(Ts_Single_Cummax(self.close,5),self.close)
        
    def bias_in45(self):
        return Cs_Double_Divide(Ts_Single_Cummax(self.close,10),self.close)
        
    def bias_in46(self):
        return Cs_Double_Divide(Ts_Single_Cummax(self.close,20),self.close)
        
    def bias_in47(self):
        return Cs_Double_Divide(Ts_Single_Cummax(self.close,30),self.close)
        
    def bias_in48(self):
        return Cs_Double_Divide(Ts_Single_Cummax(self.close,60),self.close)
        
    def bias_in49(self):
        return Cs_Double_Divide(Ts_Single_Cummin(self.close,5),self.close)
        
    def bias_in50(self):
        return Cs_Double_Divide(Ts_Single_Cummin(self.close,10),self.close)
        
    def bias_in51(self):
        return Cs_Double_Divide(Ts_Single_Cummin(self.close,20),self.close)
        
    def bias_in52(self):
        return Cs_Double_Divide(Ts_Single_Cummin(self.close,30),self.close)
        
    def bias_in53(self):
        return Cs_Double_Divide(Ts_Single_Cummin(self.close,60),self.close)
        
    def bias_in54(self):
        return Cs_Double_Divide(self.close,Ts_Single_Delay(Ts_Single_UpBoll(self.close,10,2),3))
        
    def bias_in55(self):
        return Cs_Double_Divide(self.close,Ts_Single_Delay(Ts_Single_UpBoll(self.close,30,2),3))
        
    def bias_in56(self):
        return Cs_Double_Divide(self.close,Ts_Single_Delay(Ts_Single_UpBoll(self.close,60,2),3))
        
    def bias_in57(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_DownBoll(self.close,10,2),3),self.close)
        
    def bias_in58(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_DownBoll(self.close,30,2),3),self.close)
        
    def bias_in59(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_DownBoll(self.close,60,2),3),self.close)
        
    def stability_in1(self):
        return Ts_Single_Sharp(self.pct_chg,10)
        
    def stability_in2(self):
        return Ts_Single_Sharp(self.pct_chg,20)
        
    def stability_in3(self):
        return Ts_Single_Sharp(self.pct_chg,30)
        
    def stability_in4(self):
        return Ts_Single_Sharp(self.pct_chg,60)
        
    def stability_in5(self):
        return Ts_Single_Sharp(self.turn,10)
        
    def stability_in6(self):
        return Ts_Single_Sharp(self.turn,20)
        
    def stability_in7(self):
        return Ts_Single_Sharp(self.turn,30)
        
    def stability_in8(self):
        return Ts_Single_Sharp(self.turn,60)
        
    def stability_in9(self):
        return Ts_Single_Sharpee(self.pct_chg,10)
        
    def stability_in10(self):
        return Ts_Single_Sharpee(self.pct_chg,20)
        
    def stability_in11(self):
        return Ts_Single_Sharpee(self.pct_chg,30)
        
    def stability_in12(self):
        return Ts_Single_Sharpee(self.pct_chg,60)
        
    def technique_in1(self):
        return Cs_Double_Divide(self.close-self.open+1,self.high-self.low+1)
        
    def technique_in2(self):
        return Cs_Double_Divide(self.close-self.low+1,self.high-self.open+1)
        
    def technique_in3(self):
        return Cs_Double_Divide(self.high-self.close+1,self.open-self.low+1)
        
    def technique_in4(self):
        return Ts_Single_AroonUp(self.close,7)/7
        
    def technique_in5(self):
        return Ts_Single_AroonUp(self.close,30)/30
        
    def technique_in6(self):
        return Ts_Single_AroonUp(self.close,60)/60
        
    def technique_in7(self):
        return Ts_Single_AroonDown(self.close,7)/7
        
    def technique_in8(self):
        return Ts_Single_AroonDown(self.close,30)/30
        
    def technique_in9(self):
        return Ts_Single_AroonDown(self.close,60)/60
        
    def technique_in10(self):
        return Cs_Double_Minus(Ts_Single_AroonUp(self.close,7),Ts_Single_AroonDown(self.close,7))/7
        
    def technique_in11(self):
        return Cs_Double_Minus(Ts_Single_AroonUp(self.close,30),Ts_Single_AroonDown(self.close,30))/30
        
    def technique_in12(self):
        return Cs_Double_Minus(Ts_Single_AroonUp(self.close,60),Ts_Single_AroonDown(self.close,60))/60
        
    def technique_in14(self):
        return Cs_Double_Divide(self.close-self.open,self.close.shift(1))
        
    def technique_in15(self):
        return Cs_Double_Divide(self.high-self.close,self.close.shift(1))
        
    def technique_in16(self):
        return Cs_Double_Divide(self.close-self.low,self.close.shift(1))
        
    def technique_in17(self):
        return Cs_Double_Divide(self.open-self.low,self.close.shift(1))
        
    def technique_in18(self):
        return Cs_Double_Divide(self.open-self.high,self.close.shift(1))
        
    def technique_in23(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_Rate(self.close,3),3),Ts_Single_Stddev(self.pct_chg,3))
        
    def technique_in24(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_Rate(self.close,5),5),Ts_Single_Stddev(self.pct_chg,5))
        
    def technique_in25(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_Rate(self.close,10),10),Ts_Single_Stddev(self.pct_chg,10))
        
    def technique_in26(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_Rate(self.close,20),20),Ts_Single_Stddev(self.pct_chg,20))
        
    def technique_in27(self):
        return Cs_Double_Divide(Ts_Single_Delay(Ts_Single_Rate(self.close,60),60),Ts_Single_Stddev(self.pct_chg,60))
        
    def technique_in28(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,3),Ts_Single_Delay(Ts_Single_Stddev(self.pct_chg,3),3))
        
    def technique_in29(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,5),Ts_Single_Delay(Ts_Single_Stddev(self.pct_chg,5),5))
        
    def technique_in30(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,10),Ts_Single_Delay(Ts_Single_Stddev(self.pct_chg,10),10))
        
    def technique_in31(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,20),Ts_Single_Delay(Ts_Single_Stddev(self.pct_chg,20),20))
        
    def technique_in32(self):
        return Cs_Double_Divide(Ts_Single_Rate(self.close,60),Ts_Single_Delay(Ts_Single_Stddev(self.pct_chg,60),60))
        
    def elasticity1(self):
        return Cs_Double_Divide(self.pct_chg,self.turn)
        
    def elasticity2(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.pct_chg,3),Ts_Single_Mean(self.turn,3))
        
    def elasticity3(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.pct_chg,5),Ts_Single_Mean(self.turn,5))
        
    def elasticity4(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.pct_chg,10),Ts_Single_Mean(self.turn,10))
        
    def elasticity5(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.pct_chg,30),Ts_Single_Mean(self.turn,30))
        
    def elasticity6(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.pct_chg,60),Ts_Single_Mean(self.turn,60))
        
    def elasticity7(self):
        return Cs_Double_Divide(self.swing,self.turn)
        
    def elasticity8(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.swing,3),Ts_Single_Mean(self.turn,3))
        
    def elasticity9(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.swing,5),Ts_Single_Mean(self.turn,5))
        
    def elasticity10(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.swing,10),Ts_Single_Mean(self.turn,10))
        
    def elasticity11(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.swing,30),Ts_Single_Mean(self.turn,30))
        
    def elasticity12(self):
        return Cs_Double_Divide(Ts_Single_Mean(self.swing,60),Ts_Single_Mean(self.turn,60))
        
    def elasticity13(self):
        return Cs_Double_Divide(Cs_Double_Highest(self.swing,Self_Abs(self.pct_chg)),self.turn)
        
    def elasticity14(self):
        return Cs_Double_Divide(Cs_Double_Highest(Ts_Single_Mean(self.swing,3),Ts_Single_Mean(Self_Abs(self.pct_chg),3)),self.turn)
        
    def elasticity15(self):
        return Cs_Double_Divide(Cs_Double_Highest(Ts_Single_Mean(self.swing,3),Ts_Single_Mean(Self_Abs(self.pct_chg),5)),self.turn)
        
    def elasticity16(self):
        return Cs_Double_Divide(Cs_Double_Highest(Ts_Single_Mean(self.swing,3),Ts_Single_Mean(Self_Abs(self.pct_chg),10)),self.turn)
        
    def elasticity17(self):
        return Cs_Double_Divide(Cs_Double_Highest(Ts_Single_Mean(self.swing,3),Ts_Single_Mean(Self_Abs(self.pct_chg),30)),self.turn)
        
    def elasticity18(self):
        return Cs_Double_Divide(Cs_Double_Highest(Ts_Single_Mean(self.swing,3),Ts_Single_Mean(Self_Abs(self.pct_chg),60)),self.turn)
    
    def ga_factor_1(self):
        return Ts_Single_Stddev(self.volume,8)
        
    def ga_factor_2(self):
        return Cs_Single_Correlation(self.open,18)
        
    def ga_factor_3(self):
        return Cs_Single_Correlation(self.high,16)
        
    def ga_factor_4(self):
        return Ts_Single_Max_Divide_Min(self.volume,9)
        
    def ga_factor_5(self):
        return Ts_Single_Max_Divide_Min(self.volume,9)
        
    def ga_factor_6(self):
        return Cs_Single_Correlation(self.amt,16)
        
    def ga_factor_7(self):
        return Ts_Double_Kurtosis(self.pct_chg,self.high,7)
        
    def ga_factor_8(self):
        return Cs_Single_Correlation(self.pct_chg,8)
        
    def ga_factor_9(self):
        return Ts_Single_Amplitude(Cs_Single_Correlation(self.pct_chg,8),11)
        
    def ga_factor_10(self):
        return Cs_Single_Correlation(self.close,13)
        
    def ga_factor_11(self):
        return Ts_Single_Amplitude(Cs_Single_Correlation(self.pct_chg,16),24)
        
    def ga_factor_12(self):
        return Ts_Single_Amplitude(self.amt,19)
        
    def ga_factor_13(self):
        return Ts_Single_Sharpee(Ts_Manhattan_Distance(self.volume, self.pct_chg,10),19)
        
    def ga_factor_14(self):
        return Cs_Single_Correlation(self.amt,24)
        
    def ga_factor_15(self):
        return Cs_Single_Correlation(self.amt,8)
        
    def ga_factor_16(self):
        return Ts_Single_Max_Minus_Min(Cs_Single_Correlation(self.pct_chg,8),9)
        
    def ga_factor_17(self):
        return Ts_Single_Max_Minus_Min(Cs_Single_Correlation(self.pct_chg,16),8)
        
    def ga_factor_18(self):
        return Cs_Single_Correlation(Cs_Double_Lowest(self.swing,Ts_Single_Sharp(self.pct_chg,28)),8)
        
    def ga_factor_19(self):
        return Cs_Single_Correlation(self.swing,9)
        
    def ga_factor_20(self):
        return Ts_Single_Max_Minus_Min(Cs_Single_Correlation(self.pct_chg,16),30)
        
    def ga_factor_21(self):
        return Ts_Single_Max_Minus_Min(Ts_Single_Max(Ts_Single_Stddev(self.amt,8),8),9)
        
    def ga_factor_22(self):
        return Ts_Single_Max(Cs_Single_Correlation(self.pct_chg,16),9)
        
    def ga_factor_23(self):
        return Cs_Single_Correlation(self.swing,30)
        
    def ga_factor_24(self):
        return Ts_Single_Max_Minus_Min(Cs_Single_Correlation(self.close,13),13)
        
    def ga_factor_25(self):
        return Cs_Single_Correlation(Self_SignPower_Inv(self.volume,15),9)
        
    def ga_factor_26(self):
        return Ts_Single_Max(Ts_Double_Skewness(self.amt,self.high,8),8)
        
    def ga_factor_27(self):
        return Cs_Double_Lowest(Ts_Single_Max_Minus_Min(Ts_Single_Max_Minus_Min(Self_SignPower_Inv(self.volume,15),16),19),Cs_Double_Lowest(Cs_Double_Lowest(self.swing,Ts_Single_Sharp(self.pct_chg,28)),Cs_DeepScale(self.amt,16)),19)
    
    # def factor_generate(self,function_name):
    #     factor_data = []
    #     factor_name = []
    #     for name in function_name:
    #         if hasattr(self, name) and callable(getattr(self, name)):
    #             try:
    #                 fac = getattr(self, name)().copy()
    #                 fac.columns = [name]
    #                 factor_data.append(fac)
    #             except:
    #                 print(f"函数 {name} 计算时出错")
    #                 continue
    #         else:
    #             print(f"函数 {name} 未找到")
    #     factor_data = pd.concat(factor_data,axis=1)
    #     return factor_data
    



def getCrossFactor(dateList,needStockList):
        high = np.zeros((len(dateList)*241,len(needStockList)))
        low = np.zeros((len(dateList)*241,len(needStockList)))
        open = np.zeros((len(dateList)*241,len(needStockList)))
        close = np.zeros((len(dateList)*241,len(needStockList)))
        amount = np.zeros((len(dateList)*241,len(needStockList)))
        volume = np.zeros((len(dateList)*241,len(needStockList)))
        turn = np.zeros((len(dateList)*241,len(needStockList)))
        adjfactor = np.zeros((len(dateList)*241,len(needStockList)))
        minute = np.zeros(len(dateList)*241)
        for i in range(len(needStockList)):
            code = 'stock_'+str(int(needStockList[i][:6]))+'.h5'
            data = pd.read_hdf('./stockBar/'+code)
            for j in range(len(dateList)):
                pos = np.where(data['tdate']==dateList[j])[0]
                if len(pos)==241:
                    high[(j*241):(j*241+241),i] = data['high'].iloc[pos]
                    low[(j*241):(j*241+241),i] = data['low'].iloc[pos]
                    open[(j*241):(j*241+241),i] = data['open'].iloc[pos]
                    close[(j*241):(j*241+241),i] = data['close'].iloc[pos]
                    amount[(j*241):(j*241+241),i] = data['amount'].iloc[pos]
                    volume[(j*241):(j*241+241),i] = data['volume'].iloc[pos]
                    turn[(j*241):(j*241+241),i] = data['to'].iloc[pos]
                    adjfactor[(j*241):(j*241+241),i] = data['adjfactor'].iloc[pos]
                    minute[(j*241):(j*241+241)] = data['tdate'].astype(float).iloc[pos]*10000+data['bar'].iloc[pos]
                else:
                    print(needStockList[i]+'在'+str(dateList[j])+'日数据不全')
        high = pd.DataFrame(high,index=minute, columns=needStockList)
        low = pd.DataFrame(low,index=minute, columns=needStockList)
        open = pd.DataFrame(open,index=minute, columns=needStockList)
        close = pd.DataFrame(close,index=minute, columns=needStockList)
        amount = pd.DataFrame(amount,index=minute, columns=needStockList)
        volume = pd.DataFrame(volume,index=minute, columns=needStockList)
        turn = pd.DataFrame(turn,index=minute, columns=needStockList)
        adjfactor = pd.DataFrame(adjfactor,index=minute, columns=needStockList)
        vwap = pd.DataFrame(amount.values/volume.values,index=close.index, columns=close.columns).fillna(0)
        crossStockData = {'high':high,'low':low,'open1':open,'close1':close,'vol':volume,'amt':amount,'turn':turn,'VWAP':vwap,'adjfactor':adjfactor}
        return crossStockData
    
if __name__ == '__main__':
    # dateList = [20170101,20170102]
    # needStockList = ['000333.SZ','600519.SH']
    # # 根据股票代码一个个取股票数据
    # baseFactor = getCrossFactor(dateList,needStockList)
    # # 计算因子值dd
    # needFactorList = pd.read_excel('FactorFunction_351.xlsx')['Label']
    # stock = AlphaLC(baseFactor)
    # dict0 = dict()
    # for i in range(len(needFactorList)):
    #     factorName = needFactorList[i]
    #     try:
    #         dict0[factorName] = eval('stock.' + factorName + '()')
    #     except:
    #         dict0[factorName] = None
    # # 把因子值转换成三维数组
    # subFactorValue = dict0[needFactorList[0]]
    # minuteList = subFactorValue.index.values
    # factorValue3D = np.zeros((len(minuteList), len(needStockList), len(needFactorList)))
    # for i in range(len(needFactorList)):
    #     factorValue3D[:, :, i] = dict0[needFactorList[i]].values
    feature_names = ['open', 'close', 'low', 'high', 'volume', 'pct_chg', 'amt', 'turn', 'vwap']
    data_df = dict()
    for name in feature_names:
        data_df[name] = pd.DataFrame(np.random.normal(0, 1, size=(252, 50)))
    alphalc = AlphaLC(data_df)
    print(alphalc.basic_in15())
    print("Hello world")



        


from pandas import *
from numpy import nan
from alphalens.tears import *
from alphalens.utils import *
import warnings

warnings.filterwarnings("ignore")


price_index = date_range(start='2015-1-10', end='2015-2-28')
price_index.name = 'date'
tickers = ['A', 'B', 'C', 'D', 'E', 'F']
data = [[1.0025**i, 1.005**i, 1.00**i, 0.995**i, 1.005**i, 1.00**i]
        for i in range(1, 51)]
prices = DataFrame(index=price_index, columns=tickers, data=data)


factor_index = date_range(start='2015-1-15', end='2015-2-13')
factor_index.name = 'date'
factor = DataFrame(index=factor_index, columns=tickers,
                   data=[[3, 4, 2, 1, nan, nan], [3, nan, nan, 1, 4, 2],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                         [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2]]) \
    .stack()
factor_groups = {'A': 'Group1', 'B': 'Group2', 'C': 'Group1', 'D': 'Group2', 'E': 'Group1', 'F': 'Group2'}

factor_data = get_clean_factor_and_forward_returns(
    factor,
    prices,
    groupby=factor_groups,
    quantiles=4,
    periods=(1, 3),
    filter_zscore=None)
print(factor_data)
create_full_tear_sheet(factor_data)


import pandas as pd
from pandas import DataFrame,Series
import numpy as np

import torch 
from torch.utils.data import TensorDataset, DataLoader
import torch.nn as nn
import torch.nn.functional as F
from math import log

from ...stockcalendar import CALENDAR_TOOL
import matplotlib.pyplot as plt


from ..util import DLmodel


trade_dates = Series(CALENDAR_TOOL.trade_date_in_range('2001-01-01','2026-01-01'),index = CALENDAR_TOOL.trade_date_in_range('2001-01-01','2026-01-01'))


class PositionalEncoding(nn.Module):
    """
    位置编码
    """
    def __init__(self, d_model, max_len=1000):
        super(PositionalEncoding, self).__init__()
        # 创建一个足够长的P
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return x
    

class ResidualBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1, downsample=None):
        super(ResidualBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        # self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        # self.bn2 = nn.BatchNorm2d(out_channels)
        self.downsample = downsample

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.relu(out)
        out = self.conv2(out)

        if self.downsample:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out
    

    
class Transformer5(nn.Module):
    """
    实现input_dim * seq_len -> output_dim 的基本的Transformer模型的架构, 不要前馈网络，直接跟编码器，让编码器直接提取数据特征
    """
    def __init__(self, input_dim, d_model, num_heads, num_layers, output_dim, dim_feedforward,drop_rate, backward):
        super(Transformer5, self).__init__()
        
        self.d_model = d_model
        self.backward = backward
        self.aft_dropout = nn.Dropout(drop_rate)
        
        self.pre_layer = nn.Linear(input_dim,d_model)
        
        # 编码器层
        self.pos_encoder = PositionalEncoding(self.d_model, max_len=500)
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=self.d_model,
            nhead=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=drop_rate,
            activation='relu',
            batch_first=True,
            # attn_weights_need=attn_weights_need
        )
        
        # 编码器
        self.encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)
        
        # 输出层
        self.fc1 = nn.Linear(backward*self.d_model, 500)
        self.fc2 = nn.Linear(500,100)
        self.fc3 = nn.Linear(100,output_dim)
        
        # self.init_weights()
        
    def forward(self, x):
        # x: (batch_size, seq_length, input_dim)
        
        ### 添加位置编码
        x = self.pre_layer(x)
        x = self.pos_encoder(x.transpose(0,1)).transpose(0,1)
        
        x = self.encoder(x)
        
        
        # 拼接最后若干时间步的数据再跟全连接网络
        out = x[:,-self.backward:,:].reshape(-1,self.backward*self.d_model)
        # 通过全连接层得到最终输出
        out = F.relu(self.fc1(out))
        out = self.aft_dropout(out)
        out = F.relu(self.fc2(out))
        out = self.fc3(out)
        
        return out
    


    
    

        
    



# from .account import sd
# from .stockcalendar import CALENDAR_TOOL,STOCKBASIC
# from .stockcalendar import pro as ts_pro
# import pandas as pd
# from pandas import Series,DataFrame
# import numpy as np
# import datetime
# import joblib
# from .utils import get_resource_path

# START_DATE = '19900101'
# END_DATE = datetime.date.today().strftime("%Y%m%d")


# def trans_date(date):
#     date = pd.to_datetime(date)
#     return datetime.datetime.strftime(date,'%Y-%m-%d')



# def qfq(data:DataFrame):
#     if data.empty:
#         return data['close']
#     cum = (data['close']/data['pre_close']).cumprod()
#     last_price = data['close'].iloc[-1]
#     return cum/cum.iloc[-1]*last_price

# def qfq_daily_bar(data:DataFrame):
#     if data.empty:
#         return data
#     cum = (data['close']/data['pre_close']).cumprod()
#     last_price = data['close'].iloc[-1]
#     qfq_close = cum/cum.iloc[-1]*last_price
#     qfq_open = data['open']/data['close']*qfq_close
#     qfq_high = data['high']/data['close']*qfq_close
#     qfq_low = data['low']/data['close']*qfq_close
#     data['close'] = qfq_close
#     data['open'] = qfq_open
#     data['high'] = qfq_high
#     data['low'] = qfq_low
#     return data    

# class pro:
    
#     @staticmethod
#     def ini_daily(assets:list,start_date=None,end_date=None):
#         if not isinstance(assets, list):
#             assets = list(assets)
#         if start_date is None:
#             start_date = START_DATE
#         if end_date is None:
#             end_date = END_DATE
#         df = sd.get_bars(security=assets,start_date=start_date,end_date=end_date,unit='1d',fields=["date","issue", "preclose","open","high","low","close","last","volume","value","adj","ret","is_limit_buy","is_limit_sell","numTrades"]).reset_index()
#         return df
    
#     @staticmethod
#     def daily(assets:list, start_date=None, end_date=None, need_turn=False, hfq=False):
#         if not isinstance(assets, list):
#             assets = list(assets)
#         if start_date is None:
#             start_date = START_DATE
#         if end_date is None:
#             end_date = END_DATE
        
#         if hfq:
#             adjust='hfq'
#         else:
#             adjust=''
#         df = sd.get_bars(security=assets,start_date=start_date,end_date=end_date,unit='1d',fields=["date","issue", "preclose","open","high","low","close","last","volume","value","adj",'is_limit_buy','is_limit_sell'],adjust=adjust).reset_index()
#         project = ['issue','date','open','close','high','low','preclose','value','volume','adj','is_limit_buy','is_limit_sell']
#         df = df[project]
#         df.columns = ['asset','date','open','close','high','low','pre_close','amount','volume','adj','is_limit_buy','is_limit_sell']
#         df = df.set_index(['asset','date']).sort_index()
#         if need_turn:
#             turnover = sd.get_turnover_rate(assets,start_date=start_date,end_date=end_date,fields=['today'])
#             turnover.columns = ['asset','date','turnover']
#             turnover = turnover.set_index(['asset','date']).sort_index()/100
#             df = pd.concat([df,turnover],axis=1)
#         return df
    
#     @staticmethod
#     def daily_qfq(assets:list,start_date=None,end_date=None,need_turn=False):
#         """有坑，有些数据点pre_close在停牌时被ffill了，导致cumprod之后值很大或者很小，麻了"""
#         if not isinstance(assets, list):
#             assets = list(assets)
#         df = pro.daily(assets,start_date,end_date,need_turn)
#         df = df.groupby(level=0,group_keys=False).apply(qfq_daily_bar)
#         return df
    
    
#     @staticmethod
#     def price(assets:list, start_date=None,end_date=None):
#         if not isinstance(assets, list):
#             assets = list(assets)
#         df = pro.daily(assets,start_date,end_date)
#         df = df.groupby(level=0,group_keys=False).apply(qfq).unstack().T.sort_index()
#         return df
    
#     @staticmethod
#     def index_daily(asset:str,start_date=None,end_date=None):
#         if start_date is None:
#             start_date = START_DATE
#         if end_date is None:
#             end_date = END_DATE
        
#         df = sd.get_index_bar(asset, start_date=trans_date(start_date), end_date=trans_date(end_date))
#         df = df[['date','close','high','low','open','volume']]
#         df = df.set_index(['date']).sort_index()
#         df.index = pd.to_datetime(df.index)
#         return df
    
#     @staticmethod
#     def stock_basic():
#         return STOCKBASIC
    
#     @staticmethod
#     def index_components(asset:str,end_date, break_date='20090101'):
#         trade_date = CALENDAR_TOOL.trade_date_in_range(end_date=end_date)
#         for date in np.flip(trade_date):
#             if date < pd.to_datetime(break_date):
#                 print('component not found, break')
#                 return []
#             comp = sd.get_index_stocks(asset,date=date)
#             if len(comp) > 20:
#                 return (date,comp)
#         return (date,comp)  
    
#     @staticmethod
#     def is_limit_buy(assets,start_date=None,end_date=None):
#         if not isinstance(assets, list):
#             assets = list(assets)
#         if start_date is None:
#             start_date = START_DATE
#         if end_date is None:
#             end_date = END_DATE
        
#         df = sd.get_bars(security=assets,start_date=start_date,end_date=end_date,unit='1d',fields=["date","issue","is_limit_buy"]).reset_index().pivot(index = 'date', columns= 'issue', values='is_limit_buy').sort_index()
#         return df
    
#     @staticmethod
#     def is_limit_sell(assets,start_date=None,end_date=None):
#         if not isinstance(assets, list):
#             assets = list(assets)
#         if start_date is None:
#             start_date = START_DATE
#         if end_date is None:
#             end_date = END_DATE
        
#         df = sd.get_bars(security=assets,start_date=start_date,end_date=end_date,unit='1d',fields=["date","issue","is_limit_sell"]).reset_index().pivot(index = 'date', columns= 'issue', values='is_limit_sell').sort_index()
#         return df
        
# class fpro:
    
#     @staticmethod
#     def info_derive(types=False,fields=False):
#         info = joblib.load(get_resource_path('info.derive.finance'))
#         if types:
#             return info.index.get_level_values(0).unique()
#         if fields:
#             return info.index.get_level_values(1)
#         return info
    
#     @staticmethod
#     def factor_derive(issue:list,start_date=None,end_date=None,fields:list=None,type:str=None):
#         info:pd.DataFrame = fpro.info_derive()
#         s_types = set(info.index.get_level_values(0).unique())
#         s_fields = set(info.index.get_level_values(1))
#         if start_date is None:
#             start_date = START_DATE
#         if end_date is None:
#             end_date = END_DATE
#         if type is not None:
#             if type not in s_types:
#                 raise KeyError('type is wrong!')
#             fields = list(info.loc[type].index)
#             df = sd.get_ricequant_factor(issue,start_date,end_date,fields)
#             return df
#         else:
#             if fields is None:
#                 raise KeyError('Without type input, you should provide fields list.')
#             for f in s_fields:
#                 if f not in s_fields:
#                     raise KeyError("field '{}' is wrong.".format(f))
#             df = sd.get_ricequant_factor(issue,start_date,end_date,fields)
#             return df    


import pandas as pd
import numpy as np
from pandas import DataFrame,Series
from datetime import datetime, timedelta

# from .dataget import sd, pro
from .stockcalendar import CALENDAR_TOOL, STOCKBASIC
from .data_statistics import stat_returns

##### 退市前90天和上市后90天，价格置空
def filter_stock_byDate(prices, filterdays_pre=90, filterdays_end=90):
    prices = prices.copy(deep=True)
    assets = prices.columns
    list_date = pd.to_datetime(STOCKBASIC['list_date'].loc[assets])
    delist_date = pd.to_datetime(STOCKBASIC['delist_date'].loc[assets].fillna('20800101'))
    for asset in assets:
        ld = list_date.loc[asset]
        dd = delist_date.loc[asset]
        ld += timedelta(filterdays_pre)
        dd -= timedelta(filterdays_end)
        prices[asset].loc[(prices.index < ld) | (prices.index > dd)] = np.nan
    return prices

### 根据权重表和0，1的买卖限制表重新调整为实际交易/持仓权重
def filter_weights_byLimit(weights, is_limit_buy, is_limit_sell):
    weights = weights.fillna(0)
    is_limit_buy = is_limit_buy.reindex(weights.index).fillna(0)
    is_limit_sell = is_limit_sell.reindex(weights.index).fillna(0)
    for i in range(len(weights.index)):
        if i != 0:
            chg = weights.iloc[i] - weights.iloc[i-1]
            lim_buy = is_limit_buy.iloc[i]
            lim_sell = is_limit_sell.iloc[i]
            chg.loc[(chg > 0) & (lim_buy>0)] = 0
            chg.loc[(chg < 0) & (lim_sell>0)] = 0
            weights.iloc[i] = weights.iloc[i-1] + chg
        else:
            chg = weights.iloc[i]
            lim_buy = is_limit_buy.iloc[i]
            lim_sell = is_limit_sell.iloc[i]
            chg.loc[(chg > 0) & (lim_buy>0)] = 0
            chg.loc[(chg < 0) & (lim_sell>0)] = 0
            weights.iloc[i] = chg
    return weights
    
    
class QuickBackTestor:
    def __init__(self, start_date, end_date, prices, filter_stock=True, is_limit_buy=None, is_limit_sell=None):
        self.start_date = start_date
        self.end_date = end_date
        if filter_stock:
            self._prices = filter_stock_byDate(prices, filterdays_pre=90,filterdays_end=90)
        else:
            self._prices = prices
        self._returns = self._prices.pct_change().fillna(0)
        self._commission = None
        self._trade_dates = CALENDAR_TOOL.trade_date_in_range(self.start_date, self.end_date)
        self._results = {"strategy_rewards": None,"weights": None, "turn":None, "strategy_feeadj_rewards":None}
        self._is_limit_buy = is_limit_buy
        self._is_limit_sell = is_limit_sell
        
    # def gen_is_limit(self,is_limit_buy=None,is_limit_sell=None):
    #     if is_limit_buy is None:
    #         self._is_limit_buy = pro.is_limit_buy(assets=self._prices.columns,start_date=self.start_date,end_date=self.end_date)
    #         print("Gen is_limit_buy succeed")
    #     else:
    #         self._is_limit_buy = is_limit_buy
    #     if is_limit_sell is None:
    #         self._is_limit_sell = pro.is_limit_sell(assets=self._prices.columns,start_date=self.start_date,end_date=self.end_date)
    #         print("Gen is_limit_sell succeed")
    #     else:
    #         self._is_limit_sell = is_limit_sell
    #     return None
        
        
    def _rewards_in_range(self, start_position, start_date, end_date):
        returns_slice = self._returns.loc[start_date:end_date].copy(deep=True)[start_position.index]
        # construct the imaginary returns:
        # short a stock is equal to long the stock that is purely negatively correlated with the prime stock
        returns_slice.iloc[:, np.where(start_position < 0)[0]] *= -1 
        start_position = np.abs(start_position)
        returns_slice.iloc[0] = 0
        pnls = (returns_slice + 1).cumprod()
        pnl = pnls.dot(start_position)
        # rewards = pnl.pct_change()
        # pnl = np.dot(pnls,start_position)
        # pnl = pd.Series(pnl,index=pnls.index)
        rewards = pnl.pct_change()
        return rewards.iloc[1:]
    
    def ave_backtest(self):
        return (1+self._returns).cumprod().mean(axis=1)
 
    def run_backtest(self, strategy: pd.DataFrame, is_limit=True, fee=0.002):
            
        if not set(strategy.loc[self.start_date:self.end_date].index).issubset(set(self._trade_dates)):
            raise ValueError("the index of strategy must be a trade date")
        
        if strategy.isna().sum().sum() > 0:
            raise ValueError("the elements of startegy should not be nan")

        
        ### 限制买卖，需要提供买卖限制表
        if is_limit:
            if self._is_limit_buy is None or self._is_limit_sell is None:
                raise ValueError("you need to provide the is_limit_sell and buy frame")

            strategy = filter_weights_byLimit(strategy,self._is_limit_buy,self._is_limit_sell)

        strategy = strategy.loc[self.start_date:self.end_date]
        if strategy.index[-1] != pd.to_datetime(self.end_date):
            strategy.loc[pd.to_datetime(self.end_date)] = 0
    
        rebalance_dates = strategy.index.to_list()
        dates = list(zip(rebalance_dates[:-1], rebalance_dates[1:]))
        results = []

        for start_date, end_date in dates:

            result = self._rewards_in_range(strategy.loc[start_date], start_date, end_date)
            results.append(result)
        self._results["strategy_rewards"] = pd.concat(results).reindex(self._trade_dates).fillna(0)
        self._results["strategy"] = strategy.drop(strategy.index[-1])
        self.turnover_rate()
        self.fee_adjust_rewards(fee)
        return self._results["strategy_rewards"]
    
    # def run_backtests(self, strategies: Iterable[pd.DataFrame]):
    #     """run multiple backtests
    #     The strategies should be aligned in the same format

    #     Parameters
    #     ----------
    #     strategies : Iterable[pd.DataFrame]
    #         _description_
    #     """
    #     _strategies = []
    #     for strategy in strategies:
    #         _strategies.append(strategy.loc[self.start_date:self.end_date])
    #     sample_strategy = _strategies[0]
    #     if not set(sample_strategy.index).issubset(set(self._trade_dates)):
    #         raise ValueError("the index of strategy must be a trade date")
        
    #     if sample_strategy.isna().sum().sum() > 0:
    #         raise ValueError("the elements of startegy should not be nan")
        
    #     rebalance_dates = sample_strategy.index.to_list()
    #     dates = list(zip(rebalance_dates[:-1], rebalance_dates[1:]))
    #     results = [] 
    #     for start_date, end_date in dates:
    #         current_strategy = pd.concat([strategy.loc[start_date] for strategy in strategies], axis=1, ignore_index=True)
    #         result = self._rewards_in_range(current_strategy, start_date, end_date)
    #         results.append(result)
    #     self._results["strategy_rewards"] = pd.concat(results).reindex(self._trade_dates).fillna(0)  
    #     return self._results["strategy_rewards"]

    def turnover_rate(self):
        weights = self._results['strategy']
        turn = np.abs(weights - weights.shift(1)).sum(axis=1).fillna(1)
        self._results['turn'] = turn
        if turn.max() > 2:
            print("the max turnover is {} in {}, check your codes.".format(turn.max(),turn.idxmax()))
        return self._results['turn']

    def reset(self):
        self._results = {"strategy_rewards": None,"weights": None, "turn":None, "strategy_fee_adj_rewards":None}
    
    def fee_adjust_rewards(self, fee=0.002):
        self._results['strategy_fee_adj_rewards'] = self._results['strategy_rewards'] - fee*self._results['turn'].reindex(self._results['strategy_rewards'].index).fillna(0)
        return self._results['strategy_fee_adj_rewards']
    
    def summary(self, benchmark=None):
        returns = pd.concat([self._results['strategy_rewards'], self._results['strategy_fee_adj_rewards']],axis=1)
        returns.columns = ['strategy','strategy_with_fee']
        stat_r = stat_returns(returns,benchmark)
        return stat_r




